"""Constants for guitk """

from __future__ import annotations

from enum import Enum


class GUITK(Enum):
    """Constants used internally by guitk"""

    ELEMENT_FRAME = "ttk.Frame"
    ELEMENT_LABEL_FRAME = "LabelFrame"
    ELEMENT_TK_FRAME = "tk.Frame"


DEFAULT_PADX = 5
DEFAULT_PADY = 5

MENU_MARKER = "Menu:"



"""Tools for displaying tool-tips.

This includes:
 * an abstract base-class for different kinds of tooltips
 * a simple text-only Tooltip class

This code is copied directly from CPython source code:
https://github.com/python/cpython/blob/3.9/Lib/idlelib/tooltip.py

It has been modified to work with guitk and use ttk widgets.

It is licensed under the same license as python (PSF Software Foundation License Version 2).

PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
--------------------------------------------

1. This LICENSE AGREEMENT is between the Python Software Foundation
("PSF"), and the Individual or Organization ("Licensee") accessing and
otherwise using this software ("Python") in source or binary form and
its associated documentation.

2. Subject to the terms and conditions of this License Agreement, PSF hereby
grants Licensee a nonexclusive, royalty-free, world-wide license to reproduce,
analyze, test, perform and/or display publicly, prepare derivative works,
distribute, and otherwise use Python alone or in any derivative version,
provided, however, that PSF's License Agreement and PSF's notice of copyright,
i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010,
2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021 Python Software Foundation;
All Rights Reserved" are retained in Python alone or in any derivative version
prepared by Licensee.

3. In the event Licensee prepares a derivative work that is based on
or incorporates Python or any part thereof, and wants to make
the derivative work available to others as provided herein, then
Licensee hereby agrees to include in any such work a brief summary of
the changes made to Python.

4. PSF is making Python available to Licensee on an "AS IS"
basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND
DISCLAIMS ANY REPRESENTATION OR WARRANTY OF MERCHANTABILITY OR FITNESS
FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT
INFRINGE ANY THIRD PARTY RIGHTS.

5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON
FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR LOSS AS
A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON,
OR ANY DERIVATIVE THEREOF, EVEN IF ADVISED OF THE POSSIBILITY THEREOF.

6. This License Agreement will automatically terminate upon a material
breach of its terms and conditions.

7. Nothing in this License Agreement shall be deemed to create any
relationship of agency, partnership, or joint venture between PSF and
Licensee.  This License Agreement does not grant permission to use PSF
trademarks or trade name in a trademark sense to endorse or promote
products or services of Licensee, or any third party.

8. By copying, installing or otherwise using Python, Licensee
agrees to be bound by the terms and conditions of this License
Agreement.
"""

from __future__ import annotations

import tkinter as tk
from tkinter import ttk


class _TooltipBase:
    """abstract base class for tooltips"""

    def __init__(self, anchor_widget):
        """Create a tooltip.

        anchor_widget: the widget next to which the tooltip will be shown

        Note that a widget will only be shown when showtip() is called.
        """
        self.anchor_widget = anchor_widget
        self.tipwindow = None

    def __del__(self):
        self.hidetip()

    def showtip(self):
        """display the tooltip"""
        if self.tipwindow:
            return
        self.tipwindow = tw = tk.Toplevel(self.anchor_widget)
        # show no border on the top level window
        tw.wm_overrideredirect(1)
        try:
            # This command is only needed and available on Tk >= 8.4.0 for OSX.
            # Without it, call tips intrude on the typing process by grabbing
            # the focus.
            tw.tk.call(
                "::tk::unsupported::MacWindowStyle",
                "style",
                tw._w,
                "help",
                "noActivates",
            )
        except tk.TclError:
            pass

        self.position_window()
        self.showcontents()
        self.tipwindow.update_idletasks()  # Needed on MacOS -- see #34275.
        self.tipwindow.lift()  # work around bug in Tk 8.5.18+ (issue #24570)

    def position_window(self):
        """(re)-set the tooltip's screen position"""
        x, y = self.get_position()
        root_x = self.anchor_widget.winfo_rootx() + x
        root_y = self.anchor_widget.winfo_rooty() + y
        self.tipwindow.wm_geometry("+%d+%d" % (root_x, root_y))

    def get_position(self):
        """choose a screen position for the tooltip"""
        # The tip window must be completely outside the anchor widget;
        # otherwise when the mouse enters the tip window we get
        # a leave event and it disappears, and then we get an enter
        # event and it reappears, and so on forever :-(
        #
        # Note: This is a simplistic implementation; sub-classes will likely
        # want to override this.
        return 20, self.anchor_widget.winfo_height() + 1

    def showcontents(self):
        """content display hook for sub-classes"""
        # See ToolTip for an example
        raise NotImplementedError

    def hidetip(self):
        """hide the tooltip"""
        # Note: This is called by __del__, so careful when overriding/extending
        tw = self.tipwindow
        self.tipwindow = None
        if tw:
            try:
                tw.destroy()
            except tk.TclError:  # pragma: no cover
                pass


class _OnHoverTooltipBase(_TooltipBase):
    """abstract base class for tooltips, with delayed on-hover display"""

    def __init__(self, anchor_widget, hover_delay=1000):
        """Create a tooltip with a mouse hover delay.

        anchor_widget: the widget next to which the tooltip will be shown
        hover_delay: time to delay before showing the tooltip, in milliseconds

        Note that a widget will only be shown when showtip() is called,
        e.g. after hovering over the anchor widget with the mouse for enough
        time.
        """
        super(_OnHoverTooltipBase, self).__init__(anchor_widget)
        self.hover_delay = hover_delay

        self._after_id = None
        self._id1 = self.anchor_widget.bind("<Enter>", self._show_event)
        self._id2 = self.anchor_widget.bind("<Leave>", self._hide_event)
        self._id3 = self.anchor_widget.bind("<Button>", self._hide_event)

    def __del__(self):
        try:
            self.anchor_widget.unbind("<Enter>", self._id1)
            self.anchor_widget.unbind("<Leave>", self._id2)  # pragma: no cover
            self.anchor_widget.unbind("<Button>", self._id3)  # pragma: no cover
        except tk.TclError:
            pass
        super(_OnHoverTooltipBase, self).__del__()

    def _show_event(self, event=None):
        """event handler to display the tooltip"""
        if self.hover_delay:
            self.schedule()
        else:
            self.showtip()

    def _hide_event(self, event=None):
        """event handler to hide the tooltip"""
        self.hidetip()

    def schedule(self):
        """schedule the future display of the tooltip"""
        self.unschedule()
        self._after_id = self.anchor_widget.after(self.hover_delay, self.showtip)

    def unschedule(self):
        """cancel the future display of the tooltip"""
        after_id = self._after_id
        self._after_id = None
        if after_id:
            self.anchor_widget.after_cancel(after_id)

    def hidetip(self):
        """hide the tooltip"""
        try:
            self.unschedule()
        except ttk.TclError:  # pragma: no cover
            pass
        super(_OnHoverTooltipBase, self).hidetip()


class Hovertip(_OnHoverTooltipBase):
    "A tooltip that pops up when a mouse hovers over an anchor widget."

    def __init__(self, anchor_widget, text, hover_delay=1000):
        """Create a text tooltip with a mouse hover delay.

        anchor_widget: the widget next to which the tooltip will be shown
        hover_delay: time to delay before showing the tooltip, in milliseconds

        Note that a widget will only be shown when showtip() is called,
        e.g. after hovering over the anchor widget with the mouse for enough
        time.
        """
        super(Hovertip, self).__init__(anchor_widget, hover_delay=hover_delay)
        self.text = text

    def showcontents(self):
        label = ttk.Label(
            self.tipwindow,
            text=self.text,
            justify=tk.LEFT,
            relief=tk.SOLID,
            borderwidth=1,
        )
        label.pack()



"""ttk Combobox widget"""

from __future__ import annotations

import tkinter.ttk as ttk
from typing import TYPE_CHECKING, Hashable

from .basewidget import BaseWidget
from .events import Event, EventCommand, EventType
from .types import CommandType, PadType, TooltipType

if TYPE_CHECKING:
    from .window import Window

__all__ = ["Combobox", "ComboBox"]

_valid_standard_attributes = {
    "class",
    "cursor",
    "style",
    "takefocus",
}

_valid_ttk_combobox_attributes = {
    "exportselection",
    "height",
    "justify",
    "postcommand",
    "state",
    "textvariable",
    "values",
    "width",
} | _valid_standard_attributes


class Combobox(BaseWidget):
    """ttk Combobox"""

    def __init__(
        self,
        key: Hashable | None = None,
        default: str | None = None,
        values: list[str] | None = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        keyrelease: bool = False,
        sticky: str | None = None,
        tooltip: TooltipType | None = None,
        command: CommandType | None = None,
        readonly: bool = False,
        autosize: bool = False,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        super().__init__(
            key=key,
            disabled=disabled,
            columnspan=columnspan,
            rowspan=rowspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            command=command,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
        )
        """
        Initialize a ttk.Combobox widget.

        Args:
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            default (str, optional): Default value. Defaults to None.
            values (list[str], optional): List of values for the combobox. Defaults to None.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to False.
            keyrelease (bool, optional): If True and events is True, emit EventType.KeyRelease event when a key is released.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            command (CommandType | None, optional): Command callback. Defaults to None.
            readonly (bool, optional): If True, widget is read-only. Defaults to False.
                If Combobox is not readonly, user can type in a value that is not in the list of values.
            autosize (bool, optional): If True, automatically set width to fit longest value. Defaults to False.
            weightx (int | None, optional): Weight of widget in X direction. Defaults to None.
            weighty (int | None, optional): Weight of widget in Y direction. Defaults to None.
            focus (bool, optional): If True, widget has focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments are passed to ttk.Checkbutton.

        Note:
            Emits EventType.ComboboxSelected event when a value is selected from the list.
            Emits EventType.ComboboxReturn event when the Return key is pressed.
            Emits EventType.KeyRelease event when a key is released (if keyrelease is True).
        """
        self.widget_type = "ttk.Combobox"
        self.key = key or "Combobox"
        self.columnspan = columnspan
        self.rowspan = rowspan
        self._readonly = readonly
        self._autosize = autosize
        self.kwargs = kwargs
        self.values = values
        self.default = default
        self.keyrelease = keyrelease

    def _create_widget(self, parent, window: Window, row, col):
        # build arg list for Combobox
        kwargs = {
            k: v for k, v in self.kwargs.items() if k in _valid_ttk_combobox_attributes
        }

        if self._autosize:
            # automatically set width, override any width value provided
            width = len(max(self.values, key=len))
            kwargs["width"] = width + 1

        self.widget = ttk.Combobox(
            parent,
            textvariable=self._value,
            values=self.values,
            **kwargs,
        )
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        # bind events
        if self.keyrelease:
            event_release = Event(self, window, self.key, EventType.KeyRelease)
            self.widget.bind("<KeyRelease>", window._make_callback(event_release))

        event_selected = Event(self, window, self.key, EventType.ComboboxSelected)
        self.widget.bind("<<ComboboxSelected>>", window._make_callback(event_selected))

        combo_return_key = Event(self, window, self.key, EventType.ComboboxReturn)
        self.widget.bind("<Return>", window._make_callback(combo_return_key))

        if self._command:
            self.events = True
            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.ComboboxSelected,
                    command=self._command,
                )
            )

        if self.default is not None:
            self.value = self.default

        if self._disabled:
            self.widget.state(["disabled"])

        if self._readonly:
            self.widget.state(["readonly"])

        return self.widget

    @property
    def combobox(self):
        """Return the Tk combobox widget"""
        return self.widget


class ComboBox(Combobox):
    """ttk Combobox"""

    pass



"""Utilities for the guitk package."""

from __future__ import annotations

import tkinter as tk
import tkinter.ttk as ttk


def scrolled_widget_factory(
    master: tk.BaseWidget,
    widget_class: tk.BaseWidget,
    vscrollbar: bool = False,
    hscrollbar: bool = False,
    **kwargs,
) -> tk.BaseWidget:
    """Create a widget that includes optional scrollbars"""

    # scrollbar code based on cpython source with edits to use ttk scrollbar:
    # https://github.com/python/cpython/blob/3.9/Lib/tkinter/scrolledtext.py

    frame = ttk.Frame(master) if vscrollbar or hscrollbar else None
    parent = frame or master
    widget = widget_class()

    widget.vbar = None
    if vscrollbar:
        widget.vbar = ttk.Scrollbar(frame)
        widget.vbar.grid(column=1, row=0, sticky="NS")
        # vbar.pack(side=tk.RIGHT, fill=tk.Y)
        kwargs["yscrollcommand"] = widget.vbar.set

    widget.hbar = None
    if hscrollbar:
        widget.hbar = ttk.Scrollbar(frame, orient=tk.HORIZONTAL)
        widget.hbar.grid(column=0, row=1, sticky="EW")
        # hbar.pack(side=tk.BOTTOM, fill=tk.X)
        kwargs["xscrollcommand"] = widget.hbar.set

    widget_class.__init__(widget, parent, **kwargs)
    widget.grid(column=0, row=0, sticky="NSEW")
    # widget.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

    if vscrollbar:
        widget.vbar["command"] = widget.yview

    if hscrollbar:
        widget.hbar["command"] = widget.xview

    if vscrollbar or hscrollbar:
        # Copy geometry methods of self.frame without overriding Widget methods -- hack!
        widget_meths = vars(widget_class).keys()
        methods = vars(tk.Pack).keys() | vars(tk.Grid).keys() | vars(tk.Place).keys()
        methods = methods.difference(widget_meths)

        for m in methods:
            if m[0] != "_" and m not in {"config", "configure"}:
                setattr(widget, m, getattr(frame, m))

    # keep track of whether widget is framed
    # so the correct geometry methods can be used
    widget._guitk_framed_widget = bool(frame)

    return widget


def load_image(file: str) -> tk.PhotoImage:
    """Load a photo image from a file and return it.

    If Pillow is installed, this will support more image formats than the default
    tkinter PhotoImage class. Pillow will be used automatically if it is installed.
    """
    try:
        from PIL import Image, ImageTk
    except ImportError:
        return tk.PhotoImage(file=file)
    else:
        image = Image.open(file)
        return ImageTk.PhotoImage(image)



""" ttk PanedWindow widget """

from __future__ import annotations

import tkinter.ttk as ttk
from typing import TYPE_CHECKING, Hashable, Literal, TypeVar

from guitk.constants import GUITK

from .events import EventCommand, EventType
from .frame import Frame, LabelFrame, _Container, _VerticalContainer
from .types import CommandType, PadType, TooltipType

if TYPE_CHECKING:
    from .window import Window

# TODO: Add remove() to PanedWindow and Pane

__all__ = [
    "HLabelPane",
    "HPane",
    "PanedWindow",
    "Panedwindow",
    "VLabelPane",
    "VPane",
]

_valid_standard_attributes = {
    "width",
    "height",
    "padding",
    "takefocus",
    "cursor",
    "style",
    "class",
}

_valid_ttk_panedwindow_attributes = {
    "height",
    "width",
    "orient",
} | _valid_standard_attributes


Window = TypeVar("Window")


class Panedwindow(_Container):
    """ttk.Panedwindow widget"""

    def __init__(
        self,
        key: Hashable | None = None,
        panes: list[HPane] | None = None,
        orient: Literal["horizontal", "vertical"] = "horizontal",
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        """Initialize a PanedWindow widget.

        Args:
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            panes: (list[Panes], optional): Panes to add to the Panedwindow. Defaults to None.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            weightx (int | None, optional): Horizontal weight. Defaults to None.
            weighty (int | None, optional): Vertical weight. Defaults to None.
            focus (bool, optional): If True, widget will have focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments are passed to ttk.Entry.
        """
        super().__init__(
            frametype=GUITK.ELEMENT_FRAME,
            key=None,
            width=None,
            height=None,
            layout=None,
            style=None,
            borderwidth=None,
            padding=0,
            relief=None,
            disabled=disabled,
            rowspan=None,
            columnspan=None,
            sticky=sticky,
            tooltip=None,
            autoframe=False,
            padx=0,
            pady=0,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
        )
        self.widget_type = "ttk.Panedwindow"
        self.key = key or "Panedwindow"
        self.columnspan = columnspan
        self.rowspan = rowspan
        self.panes = panes or []
        self.kwargs = kwargs
        self._pane_count = 0
        self.orient = orient
        self.padx = padx
        self.pady = pady
        self.tooltip = tooltip

    def _create_widget(self, parent, window: "Window", row, col):
        # Arg list for ttk.Label
        kwargs_panedwindow = {
            k: v
            for k, v in self.kwargs.items()
            if k in _valid_ttk_panedwindow_attributes
        }

        self.widget = ttk.Panedwindow(parent, orient=self.orient, **kwargs_panedwindow)
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        if self.layout:
            for row in self.layout:
                for pane in row:
                    self.add(pane)

        if self.width or self.height:
            self.widget.grid_propagate(0)

        if self._disabled:
            self.widget.state(["disabled"])

        return self.widget

    def add(self, pane: HPane):
        """Add a pane to the Panedwindow as new pane"""
        pane_ = pane._create_widget(self.widget, self.window, 0, 0)
        self.panedwindow.add(pane_, **pane.kwargs)

    def insert(self, pos, pane: HPane):
        """Insert a layout to the Panedwindow as new pane at position pos"""
        pane_ = pane._create_widget(self.widget, self.window, 0, 0)
        self.panedwindow.insert(pos, pane_, **pane.kwargs)

    @property
    def panedwindow(self):
        """Return the ttk.Panedwindow widget"""
        return self.widget


class PanedWindow(Panedwindow):
    """ttk.Panedwindow widget"""

    pass


class HPane(Frame):
    """Pane for Panedwindow widget that arranges its widgets horizontally"""

    def __init__(self, sticky: str | None = "nsew", **kwargs):
        """Initialize a pane"""

        super().__init__(
            frametype=GUITK.ELEMENT_FRAME,
            key=None,
            width=None,
            height=None,
            layout=None,
            style=None,
            borderwidth=None,
            padding=0,
            relief=None,
            disabled=False,
            rowspan=None,
            columnspan=None,
            sticky=sticky,
            tooltip=None,
            autoframe=True,
            padx=0,
            pady=0,
        )
        self.kwargs = kwargs


class VPane(_VerticalContainer, HPane):
    """Pane for Panedwindow widget that arranges its widgets vertically"""

    def __init__(self, sticky: str | None = "nsew", **kwargs):
        """Initialize a VerticalPane"""

        super().__init__(
            frametype=GUITK.ELEMENT_FRAME,
            key=None,
            width=None,
            height=None,
            layout=None,
            style=None,
            borderwidth=None,
            padding=0,
            relief=None,
            disabled=False,
            rowspan=None,
            columnspan=None,
            sticky=sticky,
            tooltip=None,
            autoframe=True,
            padx=0,
            pady=0,
        )
        self.kwargs = kwargs


class HLabelPane(LabelFrame):
    """Pane for Panedwindow widget that arranges its widgets horizontally and includes a label"""

    def __init__(self, name=None, sticky: str | None = "nsew", **kwargs):
        """Initialize a pane"""

        super().__init__(
            frametype=GUITK.ELEMENT_LABEL_FRAME,
            text=name,
            key=None,
            width=None,
            height=None,
            layout=None,
            style=None,
            borderwidth=None,
            padding=0,
            relief=None,
            disabled=False,
            rowspan=None,
            columnspan=None,
            sticky=sticky,
            tooltip=None,
            autoframe=True,
            padx=0,
            pady=0,
        )
        self.name = name
        self.kwargs = kwargs


class VLabelPane(_VerticalContainer, HLabelPane):
    """Pane for Panedwindow widget that includes a name and  arranges its widgets vertically"""

    def __init__(self, name=None, sticky: str | None = "nsew", **kwargs):
        """Initialize a VerticalLabelPane"""

        super().__init__(
            frametype=GUITK.ELEMENT_LABEL_FRAME,
            text=name,
            key=None,
            width=None,
            height=None,
            layout=None,
            style=None,
            borderwidth=None,
            padding=0,
            relief=None,
            disabled=False,
            rowspan=None,
            columnspan=None,
            sticky=sticky,
            tooltip=None,
            autoframe=True,
            padx=0,
            pady=0,
        )
        self.name = name
        self.kwargs = kwargs



""" Classes for handling tkinter.TK() """

from __future__ import annotations

import contextlib
import tkinter as tk
from tkinter import ttk
from typing import TypeVar

# create a custom type for type hinting a Window object which is defined later in this module
Window = TypeVar("Window")

__all__ = ["_TKRoot"]


class _TKRoot:
    """Singleton that returns a tkinter.TK() object; there can be only one in an app"""

    def __new__(cls, *args, **kwargs):
        """create new object or return instance of already created singleton"""
        if not hasattr(cls, "instance") or not cls.instance:
            cls.instance = super().__new__(cls)

        return cls.instance

    def __init__(self):
        if hasattr(self, "root"):
            return

        # create root object, make it invisible and withdraw it
        # all other windows will be children of this invisible root object
        root = tk.Tk()
        # root = customtkinter.CTk()
        root.attributes("-alpha", 0)
        root.withdraw()

        self.root = root
        self.first_window: bool = False
        self.windows: dict = {}
        self.mainloop_is_running: bool = False

    def register(self, window: Window):
        """Register a new child window"""
        if not self.first_window:
            self.first_window = True
        self.windows[window] = 1

    def deregister(self, window: Window):
        """De-register a new child window
        Once all children are de-registered, the root Tk object is destroyed
        """
        with contextlib.suppress(KeyError):
            del self.windows[window]
        if self.first_window and not self.windows:
            # last window
            self.root.quit()
            self.mainloop_is_running = False
            self.first_window = False

    def get_children(self, window):
        """Return child windows of parent window"""
        return [w for w in self.windows if w._parent == window.window]

    def run_mainloop(self):
        if not self.mainloop_is_running:
            self.root.mainloop()

    @property
    def theme(self) -> str:
        """Return name of ttk theme in use"""
        s = ttk.Style()
        return s.theme_use()

    @theme.setter
    def theme(self, theme_name: str) -> None:
        """Set name of ttk theme to use"""
        s = ttk.Style()
        theme_names = s.theme_names()
        if theme_name not in theme_names:
            raise ValueError(f"theme_name {theme_name} must by in {theme_names}")
        s.theme_use(theme_name)



"""HSpacer class that expands to fill space in the layout"""

from __future__ import annotations

import tkinter as tk
from typing import TYPE_CHECKING

from ._debug import debug_borderwidth, debug_relief
from .ttk_label import Label

if TYPE_CHECKING:
    from .window import Window

__all__ = ["HSpacer", "VSpacer"]


class HSpacer(Label):
    """HSpacer widget that expands to fill the horizontal space in the layout"""

    def __init__(self, rowspan=1):
        super().__init__(
            "",
            padding=0,
            disabled=True,
            sticky="nsew",
            autoframe=False,
            borderwidth=debug_borderwidth() or None,
            relief=debug_relief() or None,
            weightx=1,
            rowspan=rowspan,
        )

    def _create_widget(self, parent: tk.BaseWidget, window: Window, row: int, col: int):
        super()._create_widget(parent, window, row, col)
        # parent.grid_columnconfigure(col, weight=1)


class VSpacer(Label):
    """ "HSpacer widget that expands to fill the vertical space in the layout"""

    def __init__(self, columnspan=1):
        super().__init__(
            "",
            padding=0,
            disabled=True,
            sticky="nsew",
            autoframe=False,
            borderwidth=debug_borderwidth() or None,
            relief=debug_relief() or None,
            weighty=1,
            columnspan=columnspan,
        )

    def _create_widget(self, parent: tk.BaseWidget, window: Window, row: int, col: int):
        super()._create_widget(parent, window, row, col)
        # parent.grid_rowconfigure(row, weight=1)



"""ttk Button widgets"""

from __future__ import annotations

import pathlib
import tkinter as tk
import tkinter.ttk as ttk
from tkinter import filedialog
from typing import TYPE_CHECKING, Any, Hashable

from .basewidget import BaseWidget
from .events import Event, EventCommand, EventType
from .types import CommandType, PadType, TooltipType
from .utils import load_image

if TYPE_CHECKING:
    from .window import Window

__all__ = ["Button", "BrowseFileButton", "BrowseDirectoryButton"]

_valid_standard_attributes = {
    "class",
    "compound",
    "cursor",
    "image",
    "state",
    "style",
    "takefocus",
    "text",
    "textvariable",
    "underline",
    "width",
}

_valid_ttk_button_attributes = {
    "command",
    "default",
    "width",
} | _valid_standard_attributes


_valid_askopenfile_options = {
    "defaultextension",
    "filetypes",
    "initialfile",
    "title",
}


class Button(BaseWidget):
    """Basic button"""

    def __init__(
        self,
        text: str,
        image: str | None = None,
        key: Hashable | None = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        sticky: str | None = None,
        tooltip: TooltipType | None = None,
        command: CommandType | None = None,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        """
        Initialize a Button widget.

        Args:
            text (str): Text for the button.
            image (str | None, optional): Image for the button. Defaults to None.
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to False.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            command (CommandType | None, optional): Command callback. Defaults to None.
            weightx (int | None, optional): Weight in x direction. Defaults to None.
            weighty (int | None, optional): Weight in y direction. Defaults to None.
            focus (bool, optional): If True, widget has focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments are passed to ttk.Button.

        Note:
            Emits EventType.ButtonPress event.
        """
        super().__init__(
            key=key,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            command=command,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
        )

        self.widget_type = "ttk.Button"
        self.text = text
        self.image = image
        self.key = key or text
        self.columnspan = columnspan
        self.rowspan = rowspan
        self.tooltip = tooltip
        self.padx = padx
        self.pady = pady
        self.kwargs = kwargs

    @property
    def value(self) -> str:
        return self.widget["text"]

    @value.setter
    def value(self, text: str):
        self.widget["text"] = text

    def _create_widget(self, parent: Any, window: Window, row: int, col: int):
        """Create the ttk.Button widget"""
        event = Event(self, window, self.key, EventType.ButtonPress)

        # build arg list for Button()
        kwargs_button = {
            k: v for k, v in self.kwargs.items() if k in _valid_ttk_button_attributes
        }

        if self.image:
            self._photoimage = load_image(self.image)
            kwargs_button["image"] = self._photoimage

        self.widget = ttk.Button(
            parent,
            text=self.text,
            command=window._make_callback(event),
            **kwargs_button,
        )
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        if self._command:
            self.events = True
            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.ButtonPress,
                    command=self._command,
                )
            )

        if self._disabled:
            self.widget.state(["disabled"])

        return self.widget

    @property
    def button(self):
        """Return the Tk button widget"""
        return self.widget


class BrowseFileButton(Button):
    """Button that opens a file dialog to select a file."""

    def __init__(
        self,
        text="Browse",
        key: Hashable | None = None,
        target_key: Hashable | None = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        sticky: str | None = None,
        tooltip: TooltipType | None = None,
        filename_only: bool = False,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        """Initialize a BrowseFileButton widget.

        Args:
            text (str): Text for the button.
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            target_key (Hashable, optional): Unique key for the target widget. Defaults to None.
                If set, the target widget's value is set to the selected filename.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to False.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            filename_only (bool, optional): If True, only the filename is returned. Defaults to False.
            weightx (int | None, optional): Weight in x direction. Defaults to None.
            weighty (int | None, optional): Weight in y direction. Defaults to None.
            focus (bool, optional): If True, widget has focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments are passed to ttk.Button or filedialog.askopenfilename as appropriate.

        Note:
            Emits a EventType.BrowseFile event after the file dialog is closed.
        """
        super().__init__(
            text,
            key=key,
            disabled=disabled,
            columnspan=columnspan,
            rowspan=rowspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
        )
        self.target_key = target_key
        self.widget_type = "guitk.BrowseFileButton"
        self._filename = None
        self._filename_only = filename_only
        self.kwargs = kwargs

    def _create_widget(self, parent, window: "Window", row, col):
        kwargs_button = {
            k: v for k, v in self.kwargs.items() if k in _valid_ttk_button_attributes
        }
        self.widget = ttk.Button(
            parent, text=self.text, command=self.browse_dialog, **kwargs_button
        )
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )
        if self._disabled:
            self.widget.state(["disabled"])

        return self.widget

    @property
    def filename(self):
        return self._filename

    def browse_dialog(self):
        """Open a file dialog to select a file"""
        kwargs_options = {
            k: v for k, v in self.kwargs.items() if k in _valid_askopenfile_options
        }
        self._filename = filedialog.askopenfilename(**kwargs_options)
        if self._filename_only and self._filename:
            # only want the name, not the path
            self._filename = pathlib.Path(self._filename).name
        if self.target_key and self._filename:
            self.window[self.target_key].value = self._filename
        event = Event(self, self.window, self.key, EventType.BrowseFile)
        self.window._handle_event(event)


class BrowseDirectoryButton(Button):
    """Button that opens a file dialog to select a directory."""

    def __init__(
        self,
        text="Browse",
        key: Hashable | None = None,
        target_key: Hashable | None = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        sticky: str | None = None,
        tooltip: TooltipType | None = None,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        """
        Initialize a BrowseDirectoryButton widget.

        Args:
            text (str): Text for the button.
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            target_key (Hashable, optional): Unique key for the target widget. Defaults to None.
                If set, the target widget's value is set to the selected directory.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to False.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            weightx (int | None, optional): Weight in x direction. Defaults to None.
            weighty (int | None, optional): Weight in y direction. Defaults to None.
            focus (bool, optional): If True, widget has focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments are passed to ttk.Button or filedialog.askopenfilename as appropriate.

        Note:
            Emits a EventType.BrowseDirectory event after the file dialog is closed.
        """
        super().__init__(
            text,
            key=key,
            disabled=disabled,
            columnspan=columnspan,
            rowspan=rowspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
        )
        self.target_key = target_key
        self.widget_type = "guitk.BrowseDirectoryButton"
        self._dirname = None
        self.kwargs = kwargs

    def _create_widget(self, parent, window: "Window", row, col):
        kwargs_button = {
            k: v for k, v in self.kwargs.items() if k in _valid_ttk_button_attributes
        }
        self.widget = ttk.Button(
            parent, text=self.text, command=self.browse_dialog, **kwargs_button
        )
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )
        if self._disabled:
            self.widget.state(["disabled"])

        return self.widget

    @property
    def directory(self):
        return self._dirname

    def browse_dialog(self):
        """
        Open a file dialog to select a directory.
        """
        kwargs_options = {
            k: v for k, v in self.kwargs.items() if k in _valid_askopenfile_options
        }
        self._dirname = filedialog.askdirectory(**kwargs_options)
        if self.target_key and self._dirname:
            self.window[self.target_key].value = self._dirname
        event = Event(self, self.window, self.key, EventType.BrowseDirectory)
        self.window._handle_event(event)



"""Window class"""

from __future__ import annotations

import contextlib
import inspect
import time
import tkinter as tk
from tkinter import ttk
from typing import TYPE_CHECKING, Any, Hashable

from guitk.tkroot import _TKRoot

from ._debug import debug, debug_watch
from .basewidget import BaseWidget
from .constants import DEFAULT_PADX, DEFAULT_PADY, MENU_MARKER
from .events import Event, EventCommand, EventType
from .frame import _LayoutMixin
from .layout import push_parent
from .menu import Command, Menu, MenuBar
from .ttk_label import Label
from .types import PadType, SizeType, TooltipType


class _WindowBaseClass:
    # only needed to keep typing happy
    pass


class Window(_LayoutMixin, _WindowBaseClass):
    """Basic Window class from which all windows are derived

    Notes:
        Classes which inherit from window should implement handle_event, setup, and teardown as needed
    """

    def __init__(
        self,
        parent: tk.Tk | None = None,
        title: str | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        topmost: bool | None = None,
        autoframe: bool = False,
        theme: str | None = None,
        tooltip: TooltipType | None = None,
        modal: bool | None = None,
        size: SizeType = None,
    ):
        # call _config then subclass's config to initialize
        # layout, title, menu, etc.

        self.autoframe = autoframe

        self._config()
        self.config()

        # override any layout defaults from constructor
        self.title: str | None = title if title is not None else self.title
        self.padx: int | None = padx if padx is not None else self.padx
        self.pady: int | None = pady if pady is not None else self.pady
        self.theme: str | None = theme if theme is not None else self.theme
        self.tooltip: TooltipType | None = (
            tooltip if tooltip is not None else self.tooltip
        )
        self.modal: bool | None = modal if modal is not None else self.modal
        self.size: SizeType = size if size is not None else self.size

        self._id: int = id(self)
        self._tk: _TKRoot = _TKRoot()
        self._parent = parent or self._tk.root
        self._topmost = topmost

        self.window: tk.TopLevel = tk.Toplevel(self._parent)
        self.window.title(self.title)
        self._widgets = []
        self._widget_by_key = {}

        self._timer_events = {}
        """ timer events that have been set by bind_timer_event, stores most recent after() id for event"""

        self._timer_events_cancelled = {}
        """ timer events that have been set by bind_timer_event but then cancelled """

        self._return_value = None
        """ value returned from run() if set in quit() """

        self._root_menu = None
        """ will hold root tk.Menu widget"""

        self._radiobuttons = {}
        """ will hold group name/variable for radio buttons in the window """

        self._destroyed = False
        """ set to True when window is destroyed """

        self._mainframe = ttk.Frame(self.window, padding="3 3 12 12")
        self._mainframe.grid(column=0, row=0, sticky="nsew")
        self.window.columnconfigure(0, weight=1)
        self.window.rowconfigure(0, weight=1)

        self._tk.register(self)
        self.window.protocol(
            "WM_DELETE_WINDOW",
            self._make_callback(Event(self, self, EventType.Quit, EventType.Quit)),
        )

        if not self.layout:
            self.layout = [
                [
                    Label(
                        "Looks like you forgot to include a layout in your class definition"
                    )
                ]
            ]

        self._commands = []

        self._layout(self._mainframe, self)

        # apply theme if necessary
        if self.theme is not None:
            self._tk.theme = self.theme

        self._grid_configure_widgets()

        if self.menu:
            self._build_menu()

        if self._topmost or self.modal:
            self.window.attributes("-topmost", 1)

        if self.modal:
            windowingsystem = self.root.call("tk", "windowingsystem")
            if windowingsystem == "aqua":
                with contextlib.suppress(Exception):
                    self.root.call(
                        "::tk::unsupported::MacWindowStyle",
                        "style",
                        self._w,
                        "moveableModal",
                        "",
                    )
            if self._parent is not None and self._parent.winfo_viewable():
                self.window.transient(self._parent)
            self.window.wait_visibility()
            self.window.grab_set()

        if self.size is not None:
            size = (
                self.size
                if isinstance(self.size, str)
                else f"{self.size[0]}x{self.size[1]}"
            )
            self.window.geometry(size)

        # TODO: add geometry code to ensure window appears in good spot relative to parent

        # setup event handling
        self.events = True
        self._bind_timer_event(
            100, EventType.WindowFinishedLoading.value, EventType.WindowFinishedLoading
        )
        self._bind_event_handlers()
        self._create_setup_teardown_events()

        self.setup()
        self.root.event_generate(EventType.Setup.value)

        if self.modal:
            self.window.wait_window()

    def _grid_configure_widgets(self):
        """Apply padding to all widgets in the window"""
        # apply padding, widget padding takes precedent over window
        for widget in self._widgets:
            padx = widget.padx if widget.padx is not None else self.padx
            pady = widget.pady if widget.pady is not None else self.pady
            widget.widget.grid_configure(padx=padx, pady=pady)

    def _config(self):
        self.title = "My Window"
        """Title to display in the window's title bar """

        self.layout = []
        """Every class that inherits from Window must define it's own layout """

        self.menu: MenuBar | None = None
        """ Optionally provide a menu """

        self.padx = DEFAULT_PADX
        self.pady = DEFAULT_PADY
        """Default padding around widgets """

        self.theme = None
        """The ttk theme to use, if None, uses ttk default"""

        self.tooltip = None
        """ A callable which returns the tooltip text for a given key or a str """

        self.modal = False
        """ Set to True to create modal window """

        self.size = None
        """ Set to a tuple of (width, height) to set the window size """

        push_parent(self)

    def config(self):
        pass

    def handle_event(self, event: Event):
        """Handle event objects, inheriting classes should implement handle_event"""
        ...
        # if event.event_type == EventType.Quit:
        # self.quit(self._return_value)

    def setup(self):
        """Perform any needed setup for the window.
        Gets called immediately after __init__
        """
        pass

    def teardown(self):
        """Perform any cleanup before the window closes.
        Gets called immediately before the window is destroyed
        """
        pass

    @debug_watch
    def quit(self, return_value: Any = None):
        """Called when closing the window"""
        # set return value which is returned by run()
        self._return_value = return_value
        self._destroy()

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, value):
        self._title = value

    @property
    def geometry(self):
        return self.size

    @geometry.setter
    def geometry(self, value):
        self.size = value

    def bind_command(self, key=None, event_type=None, command=None):
        if not any([key, event_type]):
            raise ValueError("At least one of key, event_type must be specified")
        self._bind_command(
            EventCommand(widget=None, key=key, event_type=event_type, command=command)
        )

    def _bind_command(self, event_command: EventCommand):
        self._commands.append(event_command)

    def bind_timer_event(self, delay, event_name, repeat=False, command=None):
        """Create a new virtual event `event_name` that fires after `delay` ms,
        repeats every `delay` ms if repeat=True, otherwise fires once"""
        if command:
            self.bind_command(
                key=event_name, event_type=EventType.VirtualEvent, command=command
            )
        return self._bind_timer_event(delay, event_name, EventType.VirtualEvent, repeat)

    def _bind_timer_event(self, delay, event_name, event_type, repeat=False):
        # create a unique name for the timer
        timer_id = f"{event_name}_{time.time_ns()}"

        # callback that generates event and respawns the timer if repeat=True
        # when cancelling with cancel_timer_event, sometimes the call to after_cancel doesn't apparently work so check
        # if timer_id is in _timer_events_cancelled before respawning
        def _generate_event():
            self.root.event_generate(event_name)
            if repeat and timer_id not in self._timer_events_cancelled:
                self._timer_events[timer_id] = self._tk.root.after(
                    delay, _generate_event
                )

        event = Event(self, self, event_name, event_type)
        self.root.bind(event_name, self._make_callback(event))
        self._timer_events[timer_id] = self._tk.root.after(delay, _generate_event)
        return timer_id

    def _create_setup_teardown_events(self):
        """Create the setup and teardown events"""
        setup_event = Event(self, self, EventType.Setup.value, EventType.Setup)
        self.root.bind(EventType.Setup.value, self._make_callback(setup_event))

        teardown_event = Event(self, self, EventType.Teardown.value, EventType.Teardown)
        self.root.bind(EventType.Teardown.value, self._make_callback(teardown_event))

    def cancel_timer_event(self, timer_id):
        """Cancel a timer event created with bind_timer_event"""
        try:
            after_id = self._timer_events[timer_id]
            self.root.after_cancel(after_id)
            self._timer_events.pop(timer_id)
            self._timer_events_cancelled[timer_id] = after_id
        except KeyError as e:
            raise ValueError(f"Timer event {timer_id} not found") from e
        except Exception as e:
            raise e

    def _bind_event_handlers(self):
        """Bind any event handlers decorated with @on"""
        for method in self.__class__.__dict__.values():
            if hasattr(method, "_guitk_event_handlers"):
                for key, event_type in getattr(method, "_guitk_event_handlers"):
                    self.bind_command(key=key, event_type=event_type, command=method)

    def add_widget(self, widget: BaseWidget, row: int, col: int):
        """Add a widget to the window's mainframe"""
        widget._create_widget(self._mainframe, self, row, col)
        self._widgets.append(widget)
        self._widget_by_key[widget.key] = widget
        self._grid_configure_widgets()

    def remove(self, key_or_widget: Hashable | BaseWidget):
        """Remove widget from window and destroy it."""
        for idx, widget in enumerate(self._widgets):
            debug(f"{idx=} {widget=} {key_or_widget=}")
            if widget == key_or_widget or widget.key == key_or_widget:
                widget = self._widgets[idx]
                if widget.parent == self:
                    self._remove(widget)
                else:
                    widget.parent.remove(widget)
                return
        raise ValueError(f"Widget {key_or_widget} not found in Window")

    def _remove(self, widget: BaseWidget):
        """Remove widget from window and destroy it."""
        widget.widget.grid_forget()
        widget.widget.destroy()
        self._widget_by_key.pop(widget.key, None)
        self._widgets.remove(widget)
        self.window.update_idletasks()

    def _insert_widget_row_col(self, widget: BaseWidget, row: int, col: int):
        """Insert a widget into the window's mainframe after the container has been created
            Intended for use at run-time only when widgets need to be added dynamically

        Args:
            widget: (Widget) the widget to add
            row: (int) the row to insert the widget into
            col: (int) the column to insert the widget into

        Note:
            This method is included in Window so that Widget.replace() works properly for
            widgets added directly to a layout. It does not expand the layout like the similar
            method in _Container.
        """
        # TODO: fix this so it actually inserts instead of replaces
        self.add_widget(widget, row, col)

    def run(self):
        self._tk.run_mainloop()
        return self._return_value

    @property
    def root(self):
        """Return Tk root instance"""
        return self._tk.root

    @property
    def widgets(self) -> list[BaseWidget]:
        """ "Return list of all widgets belonging to the window"""
        return self._widgets

    def children(self):
        """Return child windows"""
        return self._tk.get_children(self)

    def get(self, key: Hashable) -> BaseWidget:
        """Get widget with key or raise KeyError if not found"""
        try:
            return self._widget_by_key[key]
        except KeyError as e:
            raise KeyError(f"Widget with key {key} not found") from e

    def _add_widget(self, widget: BaseWidget):
        """Dummy method to allow widgets to be added with VLayout()/HLayout()"""
        pass

    def _forget_widget(self, widget: BaseWidget):
        """Remove widget from the window's bookkeeping but don't destroy it"""
        self._widget_by_key.pop(widget.key, None)
        self._widgets.remove(widget)

    def _add_menus(self, menu: Menu, path: str | None = None):
        """Add menus to the window recursively

        Args:
            menu (Menu): the Menu object to add
            path (str, optional): the path to the menu item which is used as the key
        """
        path = f"{MENU_MARKER}{menu._label}" if path is None else path
        for m in menu:
            subpath = f"{path}|{m._label}"
            m._create_widget(menu._menu, self, subpath)
            self._widgets.append(m)
            self._widget_by_key[m.key] = m
            if isinstance(m, Menu):
                self._add_menus(m, subpath)

    def _build_menu(self):
        """Build the menu bar"""
        if self._root_menu is None:
            # create the root menu
            self.root.option_add("*tearOff", tk.FALSE)
            self._root_menu = tk.Menu(self.root)
            self.window["menu"] = self._root_menu

        for m in self.menu:
            if not isinstance(m, Menu):
                raise ValueError("self.menu items must be Menu objects")
            path = f"{MENU_MARKER}{m._label}"
            m._create_widget(self._root_menu, self, path)
            self._widgets.append(m)
            self._widget_by_key[m.key] = m
            self._add_menus(m, path)

    @debug_watch
    def _destroy(self):
        """Destroy the window and all child windows and perform cleanup"""
        if self._destroyed:
            # HACK: avoid multiple calls to _destroy which can occur if
            # the user handles the Quit event themselves
            # TODO: find a better way to handle this
            return

        # call teardown to perform any cleanup
        self.teardown()

        # generate a Teardown event
        self.root.event_generate(EventType.Teardown.value)

        # kill any child windows
        for child in self.children():
            event = Event(child, child.window, EventType.Quit, EventType.Quit)
            child.handle_event(event)
            with contextlib.suppress(Exception):
                child._destroy()

        # disable event processing in _handle_event
        self.events = False

        # disable any stdout/stderr redirection and event handling
        for widget in self._widgets:
            widget.events = False

            if "guitk.widgets.tk_text.Output" in str(type(widget)):
                widget.disable_redirect()

        if self.modal:
            self.window.grab_release()

        # cancel any timer events
        for timer_id in self._timer_events:
            with contextlib.suppress(Exception):
                after_id = self._timer_events[timer_id]
                self._tk.root.after_cancel(after_id)
        self._parent.focus_set()
        self.window.destroy()
        self._tk.deregister(self)
        self._destroyed = True

    def _make_callback(self, event):
        def _callback(*arg):
            if arg:
                event.event = arg[0]
            self._handle_event(event)

        return _callback

    @debug_watch
    def _handle_event(self, event: Event):
        """Handle events for this window"""
        # only handle events if widget has events=True; Window objects always get events
        if isinstance(event.widget, (BaseWidget, Window)) and not event.widget.events:
            return

        # filter events for this window
        if event.id != self._id:
            return

        # swallow MenuCommand events if the menu is disabled
        # if event.event_type == EventType.MenuCommand and not self.menu.enabled:

        # handle custom commands
        self._handle_commands(event)

        # call subclass handle_event
        self.handle_event(event)

        # if deleting the window, call _destroy after handle_event has had a chance to handle it
        if event.event_type == EventType.Quit:
            self._destroy()

    @debug_watch
    def _handle_commands(self, event):
        """Handle commands bound to widgets in the window"""
        for command in self._commands:
            if (
                (command.widget is None or command.widget == event.widget)
                and (command.key is None or command.key == event.key)
                and (
                    command.event_type is None or command.event_type == event.event_type
                )
            ) or command.event_type == EventType.Any:
                if hasattr(command.command, "_guitk_event_handlers"):
                    # command was decorated with @on, so it's a method of this class
                    if len(inspect.signature(command.command).parameters) == 2:
                        # command has a second argument, assume it's the event
                        command.command(self, event)
                    else:
                        command.command(self)
                else:
                    command.command()

    def __getitem__(self, key) -> BaseWidget:
        try:
            return self._widget_by_key[key]
        except KeyError as e:
            raise KeyError(f"Invalid key: no widget with key {key}") from e



""" Set of classes to simplify building a GUI app with tkinter.
    No dependencies outside python standard library.
    Published under the MIT License.
    Copyright Rhet Turnbull, 2023, all rights reserved.
"""

# TODO: add way to specify tooltip delay
# TODO: add style to all controls
# TODO: standardize value_type

from ._debug import debug, debug_watch, is_debug, set_debug
from ._on import on
from .basewidget import BaseWidget
from .containers import HGrid, HStack, VGrid, VStack
from .debugwindow import DebugWindow
from .events import Event, EventCommand, EventType
from .frame import Frame, LabelFrame
from .image import Image
from .layout import HLayout, VLayout
from .menu import Command, Menu, MenuBar, MenuSeparator
from .spacer import HSpacer, VSpacer
from .tk_text import Output, Text
from .tkroot import *
from .ttk_button import BrowseDirectoryButton, BrowseFileButton, Button
from .ttk_checkbutton import Checkbutton, CheckButton
from .ttk_combobox import Combobox, ComboBox
from .ttk_entry import Entry, LabelEntry
from .ttk_label import Label, LinkLabel, Linklabel
from .ttk_notebook import HTab, Notebook, NoteBook, VTab
from .ttk_panedwindow import (
    HLabelPane,
    HPane,
    Panedwindow,
    PanedWindow,
    VLabelPane,
    VPane,
)
from .ttk_progressbar import (
    PROGRESS_DETERMINATE,
    PROGRESS_INDETERMINATE,
    Progressbar,
    ProgressBar,
)
from .ttk_radiobutton import Radiobutton, RadioButton
from .ttk_scale import Scale
from .ttk_separator import HSeparator, VSeparator
from .ttk_spinbox import Spinbox, SpinBox
from .ttk_treeview import Listbox, ListBox, Treeview, TreeView
from .widget import Widget, widget_class_factory
from .window import Window

__version__ = "0.4.4"
__author__ = "Rhet Turnbull"

__all__ = [
    "BaseWidget",
    "BrowseDirectoryButton",
    "BrowseFileButton",
    "Button",
    "CheckButton",
    "Checkbutton",
    "ComboBox",
    "Combobox",
    "Command",
    "DebugWindow",
    "Entry",
    "Event",
    "EventCommand",
    "EventType",
    "Frame",
    "HGrid",
    "HLabelPane",
    "HLayout",
    "HPane",
    "HSeparator",
    "HSpacer",
    "HStack",
    "HTab",
    "Image",
    "Label",
    "LabelEntry",
    "LabelFrame",
    "LinkLabel",
    "Linklabel",
    "ListBox",
    "Listbox",
    "Menu",
    "MenuBar",
    "MenuSeparator",
    "NoteBook",
    "Notebook",
    "Output",
    "PROGRESS_DETERMINATE",
    "PROGRESS_INDETERMINATE",
    "PanedWindow",
    "Panedwindow",
    "ProgressBar",
    "Progressbar",
    "RadioButton",
    "Radiobutton",
    "Scale",
    "SpinBox",
    "Spinbox",
    "Text",
    "TreeView",
    "Treeview",
    "VGrid",
    "VLabelPane",
    "VLayout",
    "VPane",
    "VSeparator",
    "VSpacer",
    "VStack",
    "VTab",
    "Widget",
    "Window",
    "debug",
    "debug_watch",
    "is_debug",
    "on",
    "set_debug",
    "widget_class_factory",
]



""" ttk Radiobutton widget """

from __future__ import annotations

import tkinter as tk
import tkinter.ttk as ttk
from typing import TYPE_CHECKING, Any, Hashable, TypeVar, Union

from .basewidget import BaseWidget
from .events import Event, EventCommand, EventType
from .types import CommandType, PadType, TooltipType

if TYPE_CHECKING:
    from .window import Window

__all__ = ["Radiobutton", "RadioButton"]

_valid_standard_attributes = {
    "class",
    "compound",
    "cursor",
    "image",
    "state",
    "style",
    "takefocus",
    "textvariable",
    "underline",
    "width",
}

_valid_ttk_radiobutton_attributes = _valid_standard_attributes


Window = TypeVar("Window")

# TODO: need a better way to get radio button value
# only way now is the key but you really only need one key for the group to get the value


class Radiobutton(BaseWidget):
    """ttk.Radiobutton class"""

    def __init__(
        self,
        text: str,
        group: Hashable,
        key: Hashable | None = None,
        value: Union[int, str, None] = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        selected: bool = False,
        command: CommandType | None = None,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs: Any,
    ):
        """Initialize a ttk.Radiobutton widget

        Args:
            text (str): Text to display
            group (Hashable): Group name (must be unique and will be used as key unless a separate key is specified)
            key (Hashable, optional): Key to use for this widget. Defaults to None.
            value (Union[int, str, None], optional): Value to return when selected. Defaults to None.
            disabled (bool, optional): Whether to disable the widget. Defaults to False.
            columnspan (int, optional): Number of columns to span. Defaults to None.
            rowspan (int, optional): Number of rows to span. Defaults to None.
            padx (int, optional): Padding in x direction. Defaults to None.
            pady (int, optional): Padding in y direction. Defaults to None.
            events (bool, optional): Whether to enable events. Defaults to True.
            sticky (str, optional): Sticky direction. Defaults to None.
            tooltip (TooltipType, optional): Tooltip text. Defaults to None.
            selected (bool, optional): Whether to select this widget. Defaults to False.
            command (CommandType, optional): Command to execute when selected. Defaults to None.
            weightx (int, optional): Weight in x direction. Defaults to None.
            weighty (int, optional): Weight in y direction. Defaults to None.
            focus (bool, optional): If True, widget will have focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments are passed to ttk.Radiobutton.

        Note:
            Emits EventType.Radiobutton event.
        """
        super().__init__(
            key=key,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            command=command,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
        )
        self.widget_type = "ttk.Radiobutton"
        self.text = text
        self.group = group
        self.key = key or group
        self._radiobutton_value = value if value is not None else text
        self._value = None  # will be set in _create_widget
        self.columnspan = columnspan
        self.rowspan = rowspan
        self.selected = selected
        self.kwargs = kwargs

    def _create_widget(self, parent, window: "Window", row, col):
        # assign control variable or create it if necessary
        if self.group not in self.window._radiobuttons:
            # determine type of control variable based on value
            var_type = type(self._radiobutton_value)
            if var_type == int:
                self.window._radiobuttons[self.group] = tk.IntVar(
                    value=self._radiobutton_value
                )
            elif var_type == str:
                self.window._radiobuttons[self.group] = tk.StringVar(
                    value=self._radiobutton_value
                )
            else:
                # unsupported type
                raise ValueError("value must be str or int")

        self._value = self.window._radiobuttons[self.group]

        event = Event(self, window, self.key, EventType.Radiobutton)

        # Arg list for ttk.Radiobutton
        kwargs_radiobutton = {
            k: v
            for k, v in self.kwargs.items()
            if k in _valid_ttk_radiobutton_attributes
        }
        self.widget = ttk.Radiobutton(
            parent,
            text=self.text,
            command=window._make_callback(event),
            variable=self._value,
            value=self._radiobutton_value,
            **kwargs_radiobutton,
        )
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        if self._command:
            self.events = True
            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.Radiobutton,
                    command=self._command,
                )
            )

        if self._disabled:
            self.widget.state(["disabled"])

        if self.selected:
            self.widget.state(["selected"])
        else:
            self.widget.state(["!selected"])

        return self.widget

    @property
    def radiobutton(self):
        """Return the ttk Radiobutton widget"""
        return self.widget


class RadioButton(Radiobutton):
    """ttk.Radiobutton class"""

    pass



""" ScrolledFrame widget for tkinter.

    Based on similar class in [tkbootstrap](https://github.com/israel-dryer/ttkbootstrap)
    which is Copyright (c) 2021 Israel Dryer, licensed under MIT License.

    Original implementation: https://github.com/israel-dryer/ttkbootstrap/blob/7ab4e0790fb4da38c29ddffed7df0ce823b434fc/src/ttkbootstrap/scrolled.py#L184

    Unlike most other scrolled frame implementations for tkinter, this one does not use
    a canvas. Instead, it uses a ttk.Frame as the container and a ttk.Frame as the content,
    implementing xview and yview methods to scroll the content frame within the container.

    Note: The horizontal scrolling is currently not working correctly and is disabled.
"""

from __future__ import annotations

import tkinter as tk
from tkinter import Grid, Pack, Place, ttk
from typing import Any

from ._debug import debug


class ScrolledFrame(ttk.Frame):
    """A widget container with a vertical scrollbar.

    The ScrolledFrame fills the width of its container. The height is
    either set explicitly or is determined by the content frame's
    contents.

    This widget behaves mostly like a normal frame other than the
    exceptions stated already. Another exception is when packing it
    into a Notebook or Panedwindow. In this case, you'll need to add
    the container instead of the content frame. For example,
    `mynotebook.add(myscrolledframe.container)`.

    The scrollbar has an autohide feature that can be turned on by
    setting `autohide=True`.

    Args:
        parent: The parent widget.
        padding: The amount of padding to add around the content frame.
        vscrollbar: Whether to show the vertical scrollbar.
        autohide: Whether to autohide the scrollbar.
        height: The height of the content frame or None.
        width: The width of the content frame or None.
        scrollheight: The height of the container frame or None.
    """

    def __init__(
        self,
        parent=tk.BaseWidget,
        padding: tuple[int, int, int, int] | tuple[int, int] | int | str = 2,
        vscrollbar: bool = True,
        autohide: bool = False,
        height: int | None = None,
        width: int | None = None,
        scrollheight: int | None = None,
        **kwargs: dict[str, Any],
    ):
        # content frame container
        self.container = ttk.Frame(
            master=parent,
            relief=tk.FLAT,
            borderwidth=0,
            width=width,
            height=height,
        )
        self.container.propagate(False)

        # content frame
        super().__init__(
            master=self.container,
            padding=padding,
            width=width,
            height=height,
            **kwargs,
        )
        self.place(rely=0.0, relwidth=1.0, height=scrollheight)

        # vertical scrollbar
        if vscrollbar:
            self.vscroll = ttk.Scrollbar(
                master=self.container,
                command=self.yview,
                orient=tk.VERTICAL,
            )
            self.vscroll.pack(side=tk.RIGHT, fill=tk.Y)
        else:
            self.vscroll = None

        # horizontal scrollbar
        hscrollbar = False  # TODO: implement horizontal scrollbar
        if hscrollbar:
            self.hscroll = ttk.Scrollbar(
                master=self.container,
                command=self.xview,
                orient=tk.HORIZONTAL,
            )
            self.hscroll.pack(side=tk.BOTTOM, fill=tk.X)
        else:
            self.hscroll = None

        # get windowing system for mousewheel scrolling
        self.winsys = self.tk.call("tk", "windowingsystem")

        # setup autohide scrollbar
        self.autohide = autohide
        if self.autohide:
            self.hide_scrollbars()

        # widget event binding
        self.container.bind("<Configure>", self._on_configure, "+")
        self.container.bind("<Enter>", self._on_enter, "+")
        self.container.bind("<Leave>", self._on_leave, "+")
        self.container.bind("<Map>", self._on_map, "+")
        self.bind("<<MapChild>>", self._on_map_child, "+")
        self.bind("<Configure>", self._on_configure, "+")

        # delegate content geometry methods to container frame
        _methods = vars(Pack).keys() | vars(Grid).keys() | vars(Place).keys()
        for method in _methods:
            if any(["pack" in method, "grid" in method, "place" in method]):
                # prefix content frame methods with 'content_'
                setattr(self, f"content_{method}", getattr(self, method))
                # overwrite content frame methods from container frame
                setattr(self, method, getattr(self.container, method))

    def _resize_container(self, event=None):
        """Resize the container widget to fit the window."""

        requested_width = self.winfo_reqwidth()
        container_width = self.container.winfo_width()
        requested_height = self.winfo_reqheight()
        container_height = self.container.winfo_height()
        if requested_width != container_width or requested_height != container_height:
            # Resize the container widget
            new_width = max(container_width, requested_width)
            new_height = max(container_height, requested_height)
            self.container.configure(width=new_width, height=new_height)

    def yview(self, *args):
        """Update the vertical position of the content frame within the
        container.

        Parameters:

            *args (List[Any, ...]):
                Optional arguments passed to yview in order to move the
                content frame within the container frame.
        """
        if not self.vscroll:
            return
        if not args:
            first, _ = self.vscroll.get()
            self.yview_moveto(fraction=first)
        elif args[0] == "moveto":
            self.yview_moveto(fraction=float(args[1]))
        elif args[0] == "scroll":
            self.yview_scroll(number=int(args[1]), what=args[2])
        else:
            return

    def yview_moveto(self, fraction: float):
        """Update the vertical position of the content frame within the
        container.

        Parameters:

            fraction (float):
                The relative position of the content frame within the
                container.
        """
        base, thumb = self._ymeasures()
        if fraction < 0:
            first = 0.0
        elif (fraction + thumb) > 1:
            first = 1 - thumb
        else:
            first = fraction
        self.vscroll.set(first, first + thumb)
        self.content_place(rely=-first * base)

    def yview_scroll(self, number: int, what: str):
        """Update the vertical position of the content frame within the
        container.

        Parameters:

            number (int):
                The amount by which the content frame will be moved
                within the container frame by 'what' units.

            what (str):
                The type of units by which the number is to be interpeted.
                This parameter is currently not used and is assumed to be
                'units'.
        """
        first, _ = self.vscroll.get()
        fraction = (number / 100) + first
        self.yview_moveto(fraction)

    def xview(self, *args):
        """Update the horizontal position of the content frame within the
        container.

        Parameters:

            *args (List[Any, ...]):
                Optional arguments passed to yview in order to move the
                content frame within the container frame.
        """
        if not self.hscroll:
            return
        if not args:
            first, _ = self.hscroll.get()
            self.xview_moveto(fraction=first)
        elif args[0] == "moveto":
            self.xview_moveto(fraction=float(args[1]))
        elif args[0] == "scroll":
            self.xview_scroll(number=int(args[1]), what=args[2])
        else:
            return

    def xview_moveto(self, fraction: float):
        """Update the horizontal position of the content frame within the
        container.

        Parameters:

            fraction (float):
                The relative position of the content frame within the
                container.
        """
        base, thumb = self._xmeasures()

        if fraction < 0:
            first = 0.0
        elif (fraction + thumb) > 1:
            first = 1 - thumb
        else:
            first = fraction
        self.hscroll.set(first, first + thumb)
        self.content_place(relx=-first * base)
        debug(
            f"xview_moveto: {base=}, {thumb=} {fraction=} {first=} {first + thumb=} rely: {-first * base=}"
        )

    def xview_scroll(self, number: int, what: str):
        """Update the horizontal position of the content frame within the
        container.

        Parameters:

            number (int):
                The amount by which the content frame will be moved
                within the container frame by 'what' units.

            what (str):
                The type of units by which the number is to be interpeted.
                This parameter is currently not used and is assumed to be
                'units'.
        """
        first, _ = self.hscroll.get()
        fraction = (number / 100) + first
        self.xview_moveto(fraction)

    def _add_scroll_binding(self, parent):
        """Recursive adding of scroll binding to all descendants."""
        children = parent.winfo_children()
        for widget in [parent, *children]:
            bindings = widget.bind()
            if self.winsys.lower() == "x11":
                if "<Button-4>" in bindings or "<Button-5>" in bindings:
                    continue
                else:
                    widget.bind("<Button-4>", self._on_mousewheel, "+")
                    widget.bind("<Button-5>", self._on_mousewheel, "+")
            else:
                if "<MouseWheel>" not in bindings:
                    widget.bind("<MouseWheel>", self._on_mousewheel, "+")
            if widget.winfo_children() and widget != parent:
                self._add_scroll_binding(widget)

    def _del_scroll_binding(self, parent):
        """Recursive removal of scrolling binding for all descendants"""
        children = parent.winfo_children()
        for widget in [parent, *children]:
            if self.winsys.lower() == "x11":
                widget.unbind("<Button-4>")
                widget.unbind("<Button-5>")
            else:
                widget.unbind("<MouseWheel>")
            if widget.winfo_children() and widget != parent:
                self._del_scroll_binding(widget)

    def enable_scrolling(self):
        """Enable mousewheel scrolling on the frame and all of its
        children."""
        self._add_scroll_binding(self)

    def disable_scrolling(self):
        """Disable mousewheel scrolling on the frame and all of its
        children."""
        self._del_scroll_binding(self)

    def hide_scrollbars(self):
        """Hide the scrollbars."""
        if self.vscroll:
            self.vscroll.pack_forget()
        if self.hscroll:
            self.hscroll.pack_forget()

    def show_scrollbars(self):
        """Show the scrollbars."""
        if self.vscroll:
            self.vscroll.pack(side=tk.RIGHT, fill=tk.Y)
        if self.hscroll:
            self.hscroll.pack(side=tk.BOTTOM, fill=tk.X)

    def autohide_scrollbar(self):
        """Toggle the autohide funtionality. Show the scrollbars when
        the mouse enters the widget frame, and hide when it leaves the
        frame."""
        self.autohide = not self.autohide

    def _ymeasures(self):
        """Measure the base size of the container and the thumb size
        for use in the yview methods"""
        outer = self.container.winfo_height()
        inner = max([self.winfo_height(), outer])
        base = inner / outer
        if inner == outer:
            thumb = 1.0
        else:
            thumb = outer / inner
        debug(f"_ymeasures {base=} {thumb=}")
        return base, thumb

    def _xmeasures(self):
        """Calculate the size of the horizontal scrollbar thumb and base as fraction 0 to 1."""
        total_width = self.container.winfo_reqwidth()
        visible_width = self.winfo_width()
        if visible_width >= total_width:
            return 0, 1.0
        else:
            # thumb_width = max(int(visible_width / total_width * visible_width), 1)
            thumb_width = visible_width / total_width
            base_width = total_width / visible_width
            debug(f"_xmeasures {base_width=} {thumb_width=}")
            return base_width, thumb_width

    def _on_map_child(self, event):
        """Callback for when a widget is mapped to the content frame."""
        if self.container.winfo_ismapped():
            self.yview()
            self.xview()

    def _on_enter(self, event):
        """Callback for when the mouse enters the widget."""
        self.enable_scrolling()
        if self.autohide:
            self.show_scrollbars()

    def _on_leave(self, event):
        """Callback for when the mouse leaves the widget."""
        self.disable_scrolling()
        if self.autohide:
            self.hide_scrollbars()

    def _on_configure(self, event):
        """Callback for when the widget is configured"""
        self._resize_container()
        self.yview()
        self.xview()

    def _on_map(self, event):
        self._resize_container()
        self.yview()
        self.xview()

    def _on_mousewheel(self, event):
        """Callback for when the mouse wheel is scrolled."""
        if self.winsys.lower() == "win32":
            delta = -int(event.delta / 120)
        elif self.winsys.lower() == "aqua":
            delta = -event.delta
        elif event.num == 4:
            delta = -10
        elif event.num == 5:
            delta = 10
        self.yview_scroll(delta, tk.UNITS)



""" ttk Treeview and Listbox widgets """


from __future__ import annotations

import tkinter.ttk as ttk
from typing import TYPE_CHECKING, Hashable

from .basewidget import BaseWidget
from .events import Event, EventCommand, EventType
from .types import CommandType, PadType, TooltipType
from .utils import scrolled_widget_factory

if TYPE_CHECKING:
    from .window import Window

__all__ = ["ListBox", "Listbox", "TreeView", "Treeview"]

_valid_standard_attributes = {
    "class",
    "cursor",
    "displaycolumns",
    "padding",
    "selectmode",
    "show",
    "style",
    "takefocus",
    "xscrollcommand",
    "yscrollcommand",
}

_valid_ttk_treeview_attributes = _valid_standard_attributes


class Treeview(BaseWidget):
    """ttk.Treeview widget"""

    def __init__(
        self,
        headings: list[str],
        key: Hashable | None = None,
        columns: list[str] | None = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        command: CommandType | None = None,
        hscrollbar: bool = False,
        vscrollbar: bool = False,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        """Initialize a ttk.Treeview widget

        Args:
            headings (list[str]): List of column headings, required.
            key (Hashable, optional): Key to use for this widget. Defaults to None.
            columns (list[str], optional): List of column names. Defaults to None.
            disabled (bool, optional): Whether the widget is disabled. Defaults to False.
            columnspan (int, optional): Number of columns to span. Defaults to None.
            rowspan (int, optional): Number of rows to span. Defaults to None.
            padx (int, optional): Padding in x direction. Defaults to None.
            pady (int, optional): Padding in y direction. Defaults to None.
            events (bool, optional): Whether to bind events. Defaults to True.
            sticky (str, optional): Sticky direction. Defaults to None.
            tooltip (TooltipType, optional): Tooltip to display. Defaults to None.
            command (CommandType, optional): Command to run when selection changes. Defaults to None.
            hscrollbar (bool, optional): Whether to display a horizontal scrollbar. Defaults to False.
            vscrollbar (bool, optional): Whether to display a vertical scrollbar. Defaults to False.
            weightx (int, optional): Horizontal weight. Defaults to None.
            weighty (int, optional): Vertical weight. Defaults to None.
            focus (bool, optional): If True, widget will have focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments to pass to ttk.Treeview.

        Note:
            Emits EventType.TreeviewSelect event when selection changes.
            If bind_heading() is used, emits EventType.TreeviewHeading event when column heading is clicked.
            If bind_tag() is used, emits EventType.TreeviewTag event when tag is clicked.
        """
        super().__init__(
            key=key,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            command=command,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
        )
        self.key = key or "Treeview"
        self.widget_type = "ttk.Treeview"

        if headings and columns and len(headings) != len(columns):
            raise ValueError("headings and columns lists must be the same length")

        self._headings = headings

        self._columns = tuple(columns) if columns is not None else tuple(headings)
        self._disabled = disabled
        self.columnspan = columnspan
        self.rowspan = rowspan
        self.padx = padx
        self.pady = pady
        self.events = events
        self.sticky = sticky or ""
        self.tooltip = tooltip
        self.vscrollbar = vscrollbar
        self.hscrollbar = hscrollbar
        self.kwargs = kwargs

    def _create_widget(self, parent, window: Window, row, col):
        # build arg list for Treeview()
        kwargs_treeview = {
            k: v
            for k, v in self.kwargs.items()
            if k in _valid_ttk_treeview_attributes and v is not None
        }
        if "show" not in kwargs_treeview:
            kwargs_treeview["show"] = "headings"

        self.widget = scrolled_widget_factory(
            parent,
            ttk.Treeview,
            vscrollbar=self.vscrollbar,
            hscrollbar=self.hscrollbar,
            columns=self._columns,
            **kwargs_treeview,
        )
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        # set column headings
        if self._headings:
            for column, heading in zip(self._columns, self._headings):
                self.widget.heading(column, text=heading)

        if self._disabled:
            self.widget.state(["disabled"])

        event = Event(self, window, self.key, EventType.TreeviewSelect)
        self.widget.bind("<<TreeviewSelect>>", window._make_callback(event))

        if self._command:
            self.events = True
            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.TreeviewSelect,
                    command=self._command,
                )
            )

        return self.widget

    @property
    def value(self):
        return self.tree.selection()

    @value.setter
    def value(self, *values):
        self.tree.selection_set(*values)

    def bind_heading(self, column_name, event_name, command=None):
        """Bind event to click on column heading"""
        event = Event(self, self.window, event_name, EventType.TreeviewHeading)
        self.tree.heading(column_name, command=self.window._make_callback(event))

        if command:
            self.window.bind_command(
                key=event.key, event_type=event.event_type, command=command
            )

    def bind_tag(self, tagname, event_name, sequence=None, command=None):
        """Bind event to item with tag when sequence occurs
        If sequence is None, will bind to <Button-1>"""
        if sequence is None:
            sequence = "<Button-1>"
        event = Event(self, self.window, event_name, EventType.TreeviewTag)
        self.tree.tag_bind(
            tagname, sequence=sequence, callback=self.window._make_callback(event)
        )

        if command:
            self.window.bind_command(
                key=event.key, event_type=event.event_type, command=command
            )

    def sort_on_column(self, column_name, key=None, reverse=False):
        """sort the tree view contents based on column_name
        optional key same as sort(key=)
        """
        values = [
            (self.tree.set(k, column_name), k) for k in self.tree.get_children("")
        ]
        values.sort(key=key, reverse=reverse)
        for index, (val, k) in enumerate(values):
            self.tree.move(k, "", index)

    @property
    def tree(self):
        """Return the ttk Treeview widget"""
        return self.widget


class Listbox(Treeview):
    """Listbox widget (which is a Treeview widget with only one column)"""

    def __init__(
        self,
        text: list[str] | None = None,
        key: Hashable | None = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        command: CommandType | None = None,
        hscrollbar: bool = False,
        vscrollbar: bool = False,
        width: int = 200,
        weightx: int | None = None,
        weighty: int | None = None,
        **kwargs,
    ):
        """Create a Listbox widget

        Args:
            text (list[str], optional): List of strings to display in the listbox. Defaults to None.
            key (Hashable, optional): The key used to access the widget's value. Defaults to None.
            disabled (bool, optional): Disable the widget. Defaults to False.
            columnspan (int, optional): Number of columns to span in the grid. Defaults to None.
            rowspan (int, optional): Number of rows to span in the grid. Defaults to None.
            padx (int, optional): Padding in the x direction. Defaults to None.
            pady (int, optional): Padding in the y direction. Defaults to None.
            events (bool, optional): Enable events. Defaults to True.
            sticky (str, optional): Sticky setting for the grid. Defaults to None.
            tooltip (TooltipType, optional): A tooltip to show when the mouse is over the widget. Defaults to None.
            command (CommandType, optional): A command to execute when the widget is clicked. Defaults to None.
            hscrollbar (bool, optional): Show a horizontal scrollbar. Defaults to False.
            vscrollbar (bool, optional): Show a vertical scrollbar. Defaults to False.
            width (int, optional): Width of the widget in pixels. Defaults to 200.
            weightx (int, optional): Horizontal weight. Defaults to None.
            weighty (int, optional): Vertical weight. Defaults to None.
            **kwargs: Additional keyword arguments to pass to the widget constructor.

        Note:
            This Listbox widget is actually implemented as a Treeview widget with only one column.
            This is because Tk does not offer a themed ttk Listbox widget.
            Emits EventType.ListboxSelect event when the selection changes.
        """
        self.key = key or "Listbox"
        self.widget_type = "guitk.Listbox"

        if text and type(text) != list:
            raise ValueError("text must be a list of strings")
        self._text = text
        self.width = width

        super().__init__(
            headings=["list"],
            columns=["list"],
            show="tree",
            key=key,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            command=command,
            hscrollbar=hscrollbar,
            vscrollbar=vscrollbar,
            weightx=weightx,
            weighty=weighty,
            **kwargs,
        )

    def _create_widget(self, parent, window: Window, row, col):
        """Create the widget"""

        # build arg list for Treeview()
        kwargs_treeview = {
            k: v
            for k, v in self.kwargs.items()
            if k in _valid_ttk_treeview_attributes and v is not None
        }

        # listbox always shows tree
        kwargs_treeview["show"] = "tree"

        self.widget = scrolled_widget_factory(
            parent,
            ttk.Treeview,
            vscrollbar=self.vscrollbar,
            hscrollbar=self.hscrollbar,
            columns=["list"],
            **kwargs_treeview,
        )
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        self.tree.column("#0", width=0, minwidth=0)
        self.tree.column("#1", width=self.width)

        self.listbox = self.tree
        if self._text:
            for line in self._text:
                self.append(line)

        event = Event(self, window, self.key, EventType.ListboxSelect)
        self.widget.bind("<<TreeviewSelect>>", window._make_callback(event))

        if self._command:
            self.events = True
            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.ListboxSelect,
                    command=self._command,
                )
            )

        return self.widget

    def insert(self, index, line):
        """Insert a line into Listbox"""
        self.widget.insert("", index, iid=line, values=(line))

    def append(self, line):
        """Append a line to end of Listbox"""
        self.widget.insert("", "end", iid=line, values=(line))

    def delete(self, line):
        """Delete a line from Listbox"""
        self.widget.delete(line)


class TreeView(Treeview):
    """
    ttk.Treeview widget
    """

    pass


class ListBox(Listbox):
    """
    Listbox widget
    """

    pass



"""ttk.Separator widget"""

from __future__ import annotations

import tkinter as tk
import tkinter.ttk as ttk
from typing import TYPE_CHECKING

from .basewidget import BaseWidget
from .types import PadType

if TYPE_CHECKING:
    from .window import Window

__all__ = ["HSeparator", "VSeparator"]

_valid_ttk_separator_attributes = {
    "orient",
}


class HSeparator(BaseWidget):
    """ttk.Separator widget with horizontal orientation"""

    def __init__(
        self,
        columnspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        weightx: int | None = 1,
    ):
        """Initialize a horizontal HSeparator widget.

        Args:
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            weightx (int | None, optional): Weight of this widget in the horizontal direction. Defaults to 1.
        """
        super().__init__(
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            weightx=weightx,
        )
        self.widget_type = "ttk.Separator"
        self.sticky = "ew"

    def _create_widget(self, parent, window: Window, row, col):
        self.widget = ttk.Separator(parent, orient=tk.HORIZONTAL)
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        return self.widget

    @property
    def separator(self):
        """Return the ttk HSeparator widget"""
        return self.widget


class VSeparator(BaseWidget):
    """ttk.Separator widget with vertical orientation"""

    def __init__(
        self,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        weighty: int | None = 1,
    ):
        """Initialize a horizontal HSeparator widget.

        Args:
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            weightx (int | None, optional): Weight of this widget in the horizontal direction. Defaults to 1.
        """
        super().__init__(
            rowspan=rowspan,
            padx=padx,
            pady=pady,
            weighty=weighty,
        )
        self.widget_type = "ttk.Separator"
        self.sticky = "ns"

    def _create_widget(self, parent, window: Window, row, col):
        self.widget = ttk.Separator(parent, orient=tk.VERTICAL)
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        return self.widget

    @property
    def separator(self):
        """Return the ttk HSeparator widget"""
        return self.widget



""" ttk Spinbox widget"""

from __future__ import annotations

import tkinter as tk
import tkinter.ttk as ttk
from typing import TYPE_CHECKING, Hashable

from .basewidget import BaseWidget
from .events import Event, EventCommand, EventType
from .types import CommandType, PadType, TooltipType

if TYPE_CHECKING:
    from .window import Window

__all__ = ["Spinbox", "SpinBox"]

# Reference: https://tkdocs.com/pyref/ttk_spinbox.html and https://tkdocs.com/shipman/spinbox.html

_valid_ttk_spinbox_attributes = {
    "values",
    "from_",
    "to",
    "increment",
    "format",
    "command",
    "wrap",
    "exportselection",
    "font",
    "invalidcommand",
    "justify",
    "show",
    "state",
    "textvariable",
    "validate",
    "validatecommand",
    "width",
    "xscrollcommand",
    "foreground",
    "background",
    "takefocus",
    "cursor",
    "style",
    "class",
}


class Spinbox(BaseWidget):
    """ttk.Spinbox"""

    def __init__(
        self,
        from_value: float | None = None,
        to_value: float | None = None,
        values: tuple[str] | None = None,
        increment: float | None = None,
        wrap: bool = False,
        key: Hashable = None,
        target_key: Hashable = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        command: CommandType | None = None,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        """Initialize a ttk.Spinbox widget

        Args:
            from_value (float): Minimum value for the spinbox. See also values.
            to_value (float): Maximum value for the spinbox. See also values.
            values: (tuple[str], optional): Tuple of strings for the spinbox values. Defaults to None.
            increment (float, optional): Increment between from_value and to_value. Defaults to None (which is equivalent to 1.0).
            wrap (bool, optional): If True, wrap around from_value to to_value. Defaults to False.
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            target_key: (Hashable, optional): Unique key for the target widget. Defaults to None.
            default (str | None, optional): Default text for the entry box. Defaults to None.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to True.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            command (CommandType | None, optional): Command callback. Defaults to None.
            hscrollbar (bool, optional): Show horizontal scrollbar. Defaults to False.
            weightx (int | None, optional): Weight for horizontal resizing. Defaults to None.
            weighty (int | None, optional): Weight for vertical resizing. Defaults to None.
            focus (bool, optional): If True, widget has focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments are passed to ttk.Entry.

        Note:
            There are two ways to specify the possible values of the widget.
            One way is to provide a tuple of strings as the value of the values option.
            For example, values=('red', 'blue', 'green') would allow only those three strings as values.
            To configure the widget to accept a range of numeric values, see from_value, to_value, and interval.

            Emits the following events:
                EventType.SpinboxUpdate: When the spinbox value is updated.
                EventType.SpinboxDecrement: When the spinbox value is decremented.
                EventType.SpinboxIncrement: When the spinbox value is incremented.
        """

        if values and (from_value or to_value):
            raise ValueError(
                "Spinbox: Cannot specify both values and from_value/to_value"
            )

        super().__init__(
            key=key,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            command=command,
            value_type=tk.DoubleVar,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
        )
        self.widget_type = "ttk.Spinbox"
        self.key = key or "Spinbox"
        self.target_key = target_key
        self.kwargs = kwargs

        self.from_ = from_value
        self.to = to_value
        self.increment = increment
        self.values = values
        self.wrap = wrap

    @property
    def value(self):
        return self.widget.get()

    @value.setter
    def value(self, value):
        self.widget.set(value)

    def _create_widget(self, parent, window: "Window", row, col):
        """Create the ttk.Spinbox widget"""

        self.window = window
        event = Event(self, window, self.key, EventType.SpinboxUpdate)

        # build arg list for ttk.Spinbox
        kwargs = {
            k: v for k, v in self.kwargs.items() if k in _valid_ttk_spinbox_attributes
        }

        self.widget = ttk.Spinbox(
            parent,
            from_=self.from_,
            to=self.to,
            values=self.values,
            wrap=self.wrap,
            increment=self.increment,
            command=window._make_callback(event),
            **kwargs,
        )
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        # bind <<Decrement>> and <<Increment>> events
        event_decrement = Event(self, window, self.key, EventType.SpinboxDecrement)
        self.widget.bind("<<Decrement>>", window._make_callback(event_decrement))
        event_increment = Event(self, window, self.key, EventType.SpinboxIncrement)
        self.widget.bind("<<Increment>>", window._make_callback(event_increment))

        if self._command:
            self.events = True
            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.SpinboxUpdate,
                    command=self._command,
                )
            )

        if self.target_key is not None:
            self.events = True

            def update_target():
                self.window[self.target_key].value = self.value

            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.SpinboxUpdate,
                    command=update_target,
                )
            )

        if self._disabled:
            self.widget.state(["disabled"])

        return self.widget

    @property
    def spinbox(self):
        """Return the ttk.Spinbox widget"""
        return self.widget


class SpinBox(Spinbox):
    """Snake case alias for Spinbox"""

    ...



"""Menu and Command classes"""

from __future__ import annotations

import contextlib
import tkinter as tk
from inspect import currentframe, getmro
from typing import TYPE_CHECKING, Hashable

from ._debug import debug, debug_watch
from .constants import MENU_MARKER
from .events import Event, EventType
from .layout import DummyParent, get_parent, pop_parent, push_parent
from .types import CommandType

if TYPE_CHECKING:
    from .window import Window

__all__ = ["Command", "Menu", "MenuBar", "MenuSeparator"]


def _map_key_binding_from_shortcut(shortcut):
    """Return a keybinding sequence given a menu command shortcut"""
    if not shortcut:
        return None
    keys = shortcut.split("+")

    key_mappings = {"Cmd": "Command", "Ctrl": "Control"}
    keybinding = []
    for k in keys:
        if k in key_mappings:
            keybinding.append(key_mappings[k])
        else:
            if len(k) == 1 and ord("A") <= ord(k) <= ord("Z"):
                k = k.lower()
            keybinding.append(k)
    return "<" + "-".join(keybinding) + ">"


class _BaseMenu:
    """Base class for Menu and MenuBar and Command"""

    def __init__(self, label: str, disabled: bool = False):
        """Init the _BaseMenu"""
        self._label = label

        # initialize variables every subclass will need
        self._disabled: bool = disabled
        self.parent: _BaseMenu | MenuBar | None = None
        self._parent: tk.Menu | None = None
        self.key: Hashable | None = None
        self.path: str | None = None

        # _menu_list holds the list of sub-items if this is a Menu or SubMenu
        self._menu_list = []

    def __init_subclass__(subclass, *args, **kwargs):
        """Ensure that all menu items are added to the parent menu or menubar"""

        # This rewrites the __init__ method of the subclass to add the menu item to the parent menu
        super().__init_subclass__(*args, **kwargs)

        @debug_watch
        def new_init(self, *args, init=subclass.__init__, **kwargs):
            init(self, *args, **kwargs)
            if subclass is type(self):
                # only do this for the bottom grandchild class
                # in the case of subclassed menus
                self.parent = get_parent()
                if isinstance(self.parent, DummyParent):
                    raise RuntimeError(
                        "Menu item must created within a MenuBar or Menu context manager"
                    )
                debug(f"{self.parent=} {self=}")
                self.parent._add_widget(self)

        subclass.__init__ = new_init

    @property
    def disabled(self) -> bool:
        return self._parent.entrycget(self._label, "state") == tk.DISABLED

    @disabled.setter
    def disabled(self, value: bool) -> None:
        debug(f"{self=} {self._label=} {value=}")
        if value:
            self._parent.entryconfigure(self._label, state=tk.DISABLED)
        else:
            self._parent.entryconfigure(self._label, state=tk.NORMAL)
        self._disabled = self._parent.entrycget(self._label, "state") == tk.DISABLED

    def _self_or_ancestor_is_disabled(self) -> bool:
        """Returns True if self or any ancestor menu is disabled"""
        if self.disabled:
            return True
        # if self.parent is None:
        #     return self.disabled
        if isinstance(self.parent, MenuBar):
            return self.disabled
        if self._parent.entrycget(self._label, "state") == tk.DISABLED:
            return True
        return self.parent._self_or_ancestor_is_disabled()


class MenuBar:
    """Menu bar manager that can be used to create a menu bar for a Window"""

    def __init__(
        self,
    ):
        """Create a new MenuBar.

        Examples:
            ```python
            import guitk as ui

            class MenuDemo(ui.Window):
                def config(self):
                    with ui.VLayout():
                        ui.Label("This window has menus!")

                    with ui.MenuBar():
                        with ui.Menu("File"):
                            ui.Command("Open...", shortcut="Ctrl+O")
                            with ui.SubMenu("Open Recent"):
                                ui.Command("File 1")
                                ui.Command("File 2")
                                ui.Command("File 3")
                            ui.MenuSeparator()
                            ui.Command("Save", key="File|Save", disabled=True)
                            ui.Command("Save As")
                        with ui.Menu("Edit", key="Edit", disabled=True):
                            ui.Command("Cut", shortcut="Ctrl+X")
                            ui.Command("Copy", shortcut="Ctrl+C")
                            ui.Command("Paste", shortcut="Ctrl+V")
            ```
        """
        self._index = 0
        self.window = None
        self._menu_list = []

        # get the caller's instance so we can set the layout
        caller_frame = currentframe().f_back
        with contextlib.suppress(IndexError, KeyError):
            first_arg = caller_frame.f_code.co_varnames[0]
            caller_instance = caller_frame.f_locals[first_arg]
            # determine if the caller is a Window
            # need to use repr() because we can't import Window here without causing a circular import
            if "guitk.window.Window" in repr(getmro(caller_instance.__class__)):
                self.window = caller_instance
                self.window.menu = self

    def _add_widget(self, widget: _BaseMenu):
        """Add a menu item to the end of the Menu"""
        self._menu_list.append(widget)

    def __enter__(self):
        push_parent(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pop_parent()
        return False

    def __iter__(self):
        self._index = 0
        return iter(self._menu_list)

    def __next__(self) -> Menu:
        if self._index >= len(self._menu_list):
            raise StopIteration
        value = self._menu_list[self._index]
        self._index += 1
        return value


class Menu(_BaseMenu):
    def __init__(
        self,
        label: str,
        underline: int | None = None,
        key: Hashable | None = None,
        disabled: bool = False,
    ) -> None:
        """Create a new Menu.

        Args:
            label (str): The label for the menu
            underline (int, optional): The index of the character to underline in the label. Defaults to None.
            key (Hashable, optional): The key to use to access the menu from parent Window.
                Defaults to  f"Menu:{label}" for top level menus and f"Menu:{parent_label|label...}" for submenus.
            disabled: Create menu in disabled state. Defaults to False.

        Note:
            If the menu is created within a MenuBar context manager, the menu will be added to the MenuBar.
            If the menu is created within a Menu context manager, the menu will be added to the parent Menu as a submenu.

        Examples:
            ```python
            import guitk as ui

            class MenuDemo(ui.Window):
                def config(self):
                    with ui.VLayout():
                        ui.Label("This window has menus!")

                    with ui.MenuBar():
                        with ui.Menu("File"):
                            ui.Command("Open...", shortcut="Ctrl+O")
                            with ui.SubMenu("Open Recent"):
                                ui.Command("File 1")
                                ui.Command("File 2")
                                ui.Command("File 3")
                            ui.MenuSeparator()
                            ui.Command("Save", key="File|Save", disabled=True)
                            ui.Command("Save As")
                        with ui.Menu("Edit", key="Edit", disabled=True):
                            ui.Command("Cut", shortcut="Ctrl+X")
                            ui.Command("Copy", shortcut="Ctrl+C")
                            ui.Command("Paste", shortcut="Ctrl+V")
            ```
        """
        super().__init__(label, disabled=disabled)
        self._menu = None
        self._underline = underline
        self.window = None
        self._index = 0
        self.key = key

    def _create_widget(self, parent: tk.Menu, window: Window, path: str | None = None):
        """Create the Menu widget and add it to the parent

        Args:
            parent: The parent widget
            window: The Window that owns this Menu
            path: The path to the parent widget, used only by subclasses
        """
        debug(
            f"Menu._create_widget: {self=} {self._label=} {parent=} {window=} {path=}"
        )
        self.window = window
        self._parent = parent
        menu = tk.Menu(parent)
        if self._underline is None:
            idx = self._label.find("&")
            if idx != -1:
                self._label = self._label.replace("&", "", 1)
                self._underline = idx
        parent.add_cascade(menu=menu, label=self._label, underline=self._underline)

        self._menu = menu
        self.key = self.key or path
        self.path = path or ""

        if self._disabled:
            self.disabled = True

    def _add_widget(self, widget):
        """Add a menu item to the end of the Menu"""
        self._menu_list.append(widget)

    def __enter__(self):
        push_parent(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pop_parent()
        return False

    def __iter__(self):
        self._index = 0
        return iter(self._menu_list)

    def __next__(self) -> Menu | Command:
        if self._index >= len(self._menu_list):
            raise StopIteration
        value = self._menu_list[self._index]
        self._index += 1
        return value


# class SubMenu(Menu):
#     """Submenu that can be added to a Menu and holds a list of Commands or SubMenus"""

#     def __init__(
#         self,
#         label: str,
#         underline: int | None = None,
#         key: Hashable | None = None,
#         disabled: bool = False,
#     ) -> None:
#         """Create a new Submenu

#         Args:
#             label (str): The label for the submenu
#             underline (int, optional): The index of the character to underline in the label. Defaults to None.
#             key (Hashable, optional): The key to use to access the menu from parent Window.
#                 Defaults to  f"Menu:{label}" for top level menus and f"Menu:{parent_label|label...}" for submenus.
#             disabled: Create submenu in disabled state. Defaults to False.

#         Examples:
#             ```python
#             import guitk as ui

#             class MenuDemo(ui.Window):
#                 def config(self):
#                     with ui.VLayout():
#                         ui.Label("This window has menus!")

#                     with ui.MenuBar():
#                         with ui.Menu("File"):
#                             ui.Command("Open...", shortcut="Ctrl+O")
#                             with ui.SubMenu("Open Recent"):
#                                 ui.Command("File 1")
#                                 ui.Command("File 2")
#                                 ui.Command("File 3")
#                             ui.MenuSeparator()
#                             ui.Command("Save", key="File|Save", disabled=True)
#                             ui.Command("Save As")
#                         with ui.Menu("Edit", key="Edit", disabled=True):
#                             ui.Command("Cut", shortcut="Ctrl+X")
#                             ui.Command("Copy", shortcut="Ctrl+C")
#                             ui.Command("Paste", shortcut="Ctrl+V")
#             ```
#         """
#         super().__init__(label, underline=underline, key=key, disabled=disabled)


class MenuSeparator(Menu):
    """Separator that adds a dividing line between menu items.

    Examples:
        ```python
        import guitk as ui

        class MenuDemo(ui.Window):
            def config(self):
                with ui.VLayout():
                    ui.Label("This window has menus!")

                with ui.MenuBar():
                    with ui.Menu("File"):
                        ui.Command("Open...", shortcut="Ctrl+O")
                        with ui.SubMenu("Open Recent"):
                            ui.Command("File 1")
                            ui.Command("File 2")
                            ui.Command("File 3")
                        ui.MenuSeparator()
                        ui.Command("Save", key="File|Save", disabled=True)
                        ui.Command("Save As")
                    with ui.Menu("Edit", key="Edit", disabled=True):
                        ui.Command("Cut", shortcut="Ctrl+X")
                        ui.Command("Copy", shortcut="Ctrl+C")
                        ui.Command("Paste", shortcut="Ctrl+V")
        ```

    """

    def __init__(self) -> None:
        """Create a new MenuSeparator"""
        super().__init__("")

    def _create_widget(self, parent: tk.Menu, window: Window, path: str | None = None):
        """Create the Separator widget and add it to the parent

        Args:
            parent: The parent widget
            window: The Window that owns this Menu
            path: The path to the parent widget, used as the key for the command
        """
        debug(f"MenuSeparator._create_widget: {self=} {parent=} {window=} {path=}")
        self._parent = parent
        self.window = window
        parent.add_separator()
        self.key = f"{path}separator({id(self)})"
        self.path = path or ""


class Command(Menu):
    def __init__(
        self,
        label: str,
        shortcut: str = None,
        key: Hashable | None = None,
        command: CommandType | None = None,
        disabled: bool = False,
    ):
        """Create a new menu command

        Args:
            label (str): The label for the menu command
            shortcut (str, optional): The shortcut for the menu command
            key (Hashable, optional): Optional key for the menu command (defaults to the path to the menu command)
            command (CommandType, optional): The command to run when the menu command is selected
            disabled (bool): Create command in disabled state. Defaults to False.

        Note:
            Emits EventType.MenuCommand when command is selected or shortcut is pressed.

        Examples:
            ```python
            import guitk as ui

            class MenuDemo(ui.Window):
                def config(self):
                    with ui.VLayout():
                        ui.Label("This window has menus!")

                    with ui.MenuBar():
                        with ui.Menu("File"):
                            ui.Command("Open...", shortcut="Ctrl+O")
                            with ui.SubMenu("Open Recent"):
                                ui.Command("File 1")
                                ui.Command("File 2")
                                ui.Command("File 3")
                            ui.MenuSeparator()
                            ui.Command("Save", key="File|Save", disabled=True)
                            ui.Command("Save As")
                        with ui.Menu("Edit", key="Edit", disabled=True):
                            ui.Command("Cut", shortcut="Ctrl+X")
                            ui.Command("Copy", shortcut="Ctrl+C")
                            ui.Command("Paste", shortcut="Ctrl+V")
            ```
        """
        super().__init__(label, disabled=disabled)
        self.shortcut = shortcut
        self._parent = None
        self.key = key
        self._command = command

    def _create_widget(self, parent: tk.Menu, window: Window, path: str):
        """Create the Menu widget and add it to the parent

        Args:
            parent: The parent widget
            window: The Window that owns this Menu
            path: The path to the parent widget, used as the key for the command
        """
        debug(f"Command._create_widget: {self=} {parent=} {window=} {path=}")
        self._parent = parent
        self.window = window
        self.key = self.key or path
        self.path = path or ""

        parent.add_command(
            label=self._label,
            command=self.window._make_callback(
                Event(self, self.window, self.key, EventType.MenuCommand)
            ),
            accelerator=self.shortcut,
        )

        if self._command:
            self.window.bind_command(
                key=self.key, event_type=EventType.MenuCommand, command=self._command
            )

        if self._disabled:
            self.disabled = True

        key_binding = _map_key_binding_from_shortcut(self.shortcut)

        def _make_callback(event):
            def _callback(*arg):
                if self._self_or_ancestor_is_disabled():
                    # if self or any ancestor is disabled, ignore the shortcut event
                    return
                if arg:
                    event.event = arg[0]
                self.window._handle_event(event)

            return _callback

        window.window.bind_all(
            key_binding,
            _make_callback(Event(self, self.window, self.key, EventType.MenuCommand)),
        )



"""@on decorator for defining event handlers """

from __future__ import annotations

from functools import wraps
from typing import Callable, Hashable

from .events import EventType
from .types import CommandType, DecoratedType


def on(
    key: Hashable | None = None,
    event_type: EventType | None = None,
) -> CommandType:
    """Decorator to declare that the method is an event handler.

    The decorated method will be called when a widget with the given `key`
    emits an event or an event of the given `event_type` is emitted.

    Example:
        ```python
        # Handle the press of buttons with key "OK"
        @on(key="OK")
        def ok_button(self) -> None:
            ...

        # Handle the press of any button
        @on(event_type=EventType.ButtonPressed)
        def any_button(self) -> None:
            ...

        # Handle the press of buttons with key "OK" or "Cancel"
        @on(key="OK", event_type=EventType.ButtonPressed)
        @on(key="Cancel", event_type=EventType.ButtonPressed)
        def ok_or_cancel_button(self) -> None:
            ...

        # Handle any event
        @on(event_type=EventType.Any)
        def any_event(self) -> None:
            ...
        ```

    Args:
        key (Hashable, optional): Key of the event to handle. Defaults to None.
        event_type (EventType, optional): Type of the event to handle. Defaults to None.

    Note: Either `key` or `event_type` must be specified and both can be specified.
    """
    # Inspired by the implementation of the `on` decorator in textual
    # https://github.com/Textualize/textual

    # Event handlers registered with the `on` decorator are stored in the
    # `_guitk_event_handlers` attribute of the decorated method and are bound
    # by Window._bind_event_handlers() then called by Window._handle_commands()

    if not key and not event_type:
        raise ValueError("Either key or event_type (or both) must be specified")

    def decorate(method: Callable[..., DecoratedType]):
        if not hasattr(method, "_guitk_event_handlers"):
            method._guitk_event_handlers = []
        getattr(method, "_guitk_event_handlers").append((key, event_type))

        @wraps(method)
        def wrapper(*args, **kwargs) -> DecoratedType:
            """Store key, event_type, return callable unaltered."""
            return method(*args, **kwargs)

        return wrapper

    return decorate



"""Events """

from __future__ import annotations

import tkinter
from collections import namedtuple
from typing import TYPE_CHECKING, Hashable

EventCommand = namedtuple("EventCommand", ["widget", "key", "event_type", "command"])

from enum import Enum

if TYPE_CHECKING:
    from .basewidget import BaseWidget
    from .window import Window


class Event:
    """Event that occurred and values for widgets in the window"""

    def __init__(
        self, widget: BaseWidget, window: Window, key: Hashable, event_type: EventType
    ):
        self.id: int = id(window)
        self.widget: BaseWidget = widget
        self.key: Hashable = key
        self.event_type: EventType = event_type
        self.event: tkinter.Event | None = (
            None  # placeholder for Tk event, will be set in _make_callback
        )

    def __str__(self):
        return f"id={self.id}, widget={self.widget}, key={self.key}, event_type={self.event_type}, event={self.event}"


class EventType(Enum):
    """Constants for event types"""

    Any = "*"  # special case for binding to all events
    BrowseDirectory = "<<BrowseDirectory>>"
    BrowseFile = "<<BrowseFile>>"
    ButtonPress = "<<Button>>"
    CheckButton = "<<Checkbutton>>"
    Checkbutton = "<<Checkbutton>>"
    ComboboxReturn = "<<ComboboxReturn>>"
    ComboBoxReturn = "<<ComboboxReturn>>"
    ComboBoxSelected = "<<ComboboxSelected>>"
    ComboboxSelected = "<<ComboboxSelected>>"
    DeleteWindow = "WM_DELETE_WINDOW"
    EntryReturn = "<<EntryReturn>>"
    ImagePress = "<<ImagePress>>"
    KeyRelease = "<KeyRelease>"
    LinkLabel = "<<LinkLabel>>"
    ListBoxSelect = "<<ListboxSelect>>"
    ListboxSelect = "<<ListboxSelect>>"
    MenuCommand = "<<MenuCommand>>"
    NotebookTabChanged = "<<NotebookTabChanged>>"
    OutputWrite = "<<OutputWrite>>"
    Quit = "WM_DELETE_WINDOW"
    RadioButton = "<<Radiobutton>>"
    Radiobutton = "<<Radiobutton>>"
    ScaleUpdate = "<<ScaleUpdate>>"
    Setup = "<<Setup>>"
    SpinboxDecrement = "<<SpinboxDecrement>>"
    SpinboxIncrement = "<<SpinboxIncrement>>"
    SpinboxUpdate = "<<SpinboxUpdate>>"
    Teardown = "<<Teardown>>"
    TreeViewHeading = "<<TreeviewHeading>>"
    TreeViewSelect = "<<TreeviewSelect>>"
    TreeViewTag = "<<TreeviewTag>>"
    TreeviewHeading = "<<TreeviewHeading>>"
    TreeviewSelect = "<<TreeviewSelect>>"
    TreeviewTag = "<<TreeviewTag>>"
    VirtualEvent = "<<VirtualEvent>>"
    WM_DELETE_WINDOW = "WM_DELETE_WINDOW"
    WindowFinishedLoading = "<<WindowFinishedLoading>>"



"""Image widget"""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Hashable

from .events import Event, EventCommand, EventType
from .ttk_label import Label
from .types import CommandType, CompoundType, PadType, TooltipType, Window

if TYPE_CHECKING:
    from .window import Window

__all__ = ["Image"]


class Image(Label):
    """Image widget"""

    def __init__(
        self,
        image: str,
        text: str | None = None,
        compound: CompoundType = None,
        key: Hashable | None = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = False,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        command: CommandType | None = None,
        weightx: int | None = None,
        weighty: int | None = None,
        **kwargs,
    ):
        """
        Initialize an Image widget.

        Args:
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            image: (str, optional): Path to image to display.
            text (str, optional): Text to display with the image.
            compound (str, optional): How to display the image and text. Defaults to None (show image only).
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to False.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            command (CommandType | None, optional): Command to execute when clicked. Defaults to None.
            weightx (int | None, optional): Weight of this widget in the horizontal direction. Defaults to None.
            weighty (int | None, optional): Weight of this widget in the vertical direction. Defaults to None.
            **kwargs: Additional keyword arguments are passed to ttk.Entry.

        Note:
            tkinter only supports images in GIF, PGM, PPM, and PNG formats.
            If you install the option Pillow package, you can use images in many other formats as well.
            guitk will automatically use Pillow if it is installed.

            compound can be one of: None, "top", "bottom", "left", "right", "center", "image", "text"
            If None or "image", only the image is displayed.
            If "text", only the text is displayed.
            If "top", "bottom", "left", or "right", the image is displayed above, below, to the left, or to the right of the text.

            Emits EventType.ImagePress when clicked if events is True.
        """
        super().__init__(
            text=text,
            key=key,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            weightx=weightx,
            weighty=weighty,
        )
        self.widget_type = "guitk.Image"
        self.image = image
        self.text = text
        self.compound = compound
        self.key = key or text
        self.columnspan = columnspan
        self.rowspan = rowspan
        self.kwargs = kwargs

        self.cursor = (
            kwargs.get("cursor") or "pointinghand"
            if sys.platform == "darwin"
            else "hand2"
        )

    def _create_widget(self, parent, window: "Window", row, col):
        self.widget = super()._create_widget(parent, window, row, col)

        if self.compound:
            self.widget.configure(compound=self.compound)

        if self.events:
            event = Event(self, window, self.key, EventType.ImagePress)
            self.widget.bind("<Button-1>", window._make_callback(event))
            self.widget.configure(cursor=self.cursor)

        if self._command:
            self.events = True
            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.LinkLabel,
                    command=self._command,
                )
            )
        return self.widget



"""ttk checkbutton widget"""

from __future__ import annotations

import tkinter as tk
import tkinter.ttk as ttk
from typing import TYPE_CHECKING, Hashable

from .basewidget import BaseWidget
from .events import Event, EventCommand, EventType
from .types import CommandType, PadType, TooltipType

if TYPE_CHECKING:
    from .window import Window

__all__ = ["Checkbutton", "CheckButton"]

_valid_standard_attributes = {
    "class",
    "compound",
    "cursor",
    "image",
    "state",
    "style",
    "takefocus",
    "text",
    "textvariable",
    "underline",
    "width",
}

_valid_ttk_checkbutton_attributes = {
    "command",
    "variable",
    # "offvalue",
    # "onvalue",
} | _valid_standard_attributes


class Checkbutton(BaseWidget):
    """Checkbox / checkbutton"""

    def __init__(
        self,
        text: str,
        key: Hashable | None = None,
        checked: bool = False,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        sticky: str | None = None,
        tooltip: TooltipType | None = None,
        command: CommandType | None = None,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        """
        Initialize a ttk.Checkbutton widget.

        Args:
            text (str): Text for the checkbutton.
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            checked (bool, optional): Initial state. Defaults to False (not checked).
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to False.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            command (CommandType | None, optional): Command callback. Defaults to None.
            weightx (int | None, optional): Weight of widget in X direction. Defaults to None.
            weighty (int | None, optional): Weight of widget in Y direction. Defaults to None.
            focus (bool, optional): If True, widget has focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments are passed to ttk.Checkbutton.

        Notes:
            Unlike a regular ttk.Checkbutton, the onvalue and offvalue are always True and False.
            Emits an EventType.Checkbutton event when the checkbutton is clicked.
        """
        super().__init__(
            key=key,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            command=command,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
        )
        self.widget_type = "ttk.Checkbutton"
        self.text = text
        self.key = key or text
        self.columnspan = columnspan
        self.rowspan = rowspan
        self._value: tk.BooleanVar = tk.BooleanVar()
        self._checked = checked
        self.kwargs = kwargs

    def _create_widget(self, parent, window: Window, row, col):
        event = Event(self, window, self.key, EventType.Checkbutton)

        # build arg list for Checkbutton
        kwargs_checkbutton = {
            k: v
            for k, v in self.kwargs.items()
            if k in _valid_ttk_checkbutton_attributes
        }

        self.widget = ttk.Checkbutton(
            parent,
            text=self.text,
            anchor=self.kwargs.get("anchor"),
            command=window._make_callback(event),
            variable=self._value,
            onvalue=True,
            **kwargs_checkbutton,
        )
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        if self._command:
            self.events = True
            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.Checkbutton,
                    command=self._command,
                )
            )

        if self._checked:
            self.widget.invoke()

        if self._disabled:
            self.widget.state(["disabled"])
        return self.widget

    @property
    def checkbutton(self):
        """Return the ttk.Checkbutton widget"""
        return self.widget


class CheckButton(Checkbutton):
    """Checkbox / checkbutton"""

    pass



""" tk Text and Output widgets """


from __future__ import annotations

import contextlib
import tkinter as tk
from typing import Hashable, TypeVar

from guitk.redirect import StdErrRedirect, StdOutRedirect

from .basewidget import BaseWidget
from .events import Event, EventCommand, EventType
from .types import CommandType, PadType, TooltipType
from .utils import scrolled_widget_factory

__all__ = ["Text", "Output"]


_valid_standard_attributes = {
    "background",
    "borderwidth",
    "cursor",
    "exportselection",
    "font",
    "foreground",
    "highlightbackground",
    "highlightcolor",
    "highlightthickness",
    "insertbackground",
    "insertborderwidth",
    "insertofftime",
    "insertontime",
    "insertwidth",
    "padx",
    "pady",
    "relief",
    "selectbackground",
    "selectborderwidth",
    "selectforeground",
    "setgrid",
    "takefocus",
    "xscrollcommand",
    "yscrollcommand",
}

_valid_tk_text_attributes = {
    "autoseparators",
    "maxundo",
    "spacing1",
    "spacing2",
    "spacing3",
    "state",
    "tabs",
    "undo",
    "wrap",
} | _valid_standard_attributes


Window = TypeVar("Window")


class Text(BaseWidget):
    """A tk Text box"""

    def __init__(
        self,
        text: str | None = None,
        key: Hashable | None = None,
        width: int = 40,
        height: int = 20,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = False,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        command: CommandType | None = None,
        vscrollbar: bool = False,
        hscrollbar: bool = False,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        """
        Initialize a Text widget.

        Args:
            text (str | None, optional): Default text for the text box. Defaults to None.
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            width (int, optional): Width of the text box. Defaults to 40.
            height (int, optional): Height of the text box. Defaults to 20.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to False.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            command (CommandType | None, optional): Command callback. Defaults to None.
            vscrollbar (bool, optional): Show vertical scrollbar. Defaults to False.
            hscrollbar (bool, optional): Show horizontal scrollbar. Defaults to False.
            weightx (int | None, optional): Weight of the widget in the x direction. Defaults to None.
            weighty (int | None, optional): Weight of the widget in the y direction. Defaults to None.
            focus (bool, optional): If True, widget has focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments are passed to tk Text.

        Note:
            Emits EventType.KeyRelease events when the text is changed and events is True.
        """
        super().__init__(
            key=key,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            command=command,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
        )
        self.widget_type = "tk.Text"
        self.key = key or "Text"
        self.width = width
        self.height = height
        self._value = text if text is not None else ""
        self.columnspan = columnspan
        self.rowspan = rowspan
        self.vscrollbar = vscrollbar
        self.hscrollbar = hscrollbar
        self.kwargs = kwargs

    def _create_widget(self, parent, window: "Window", row, col):
        kwargs_text = {
            k: v for k, v in self.kwargs.items() if k in _valid_tk_text_attributes
        }
        self.widget = scrolled_widget_factory(
            parent,
            tk.Text,
            vscrollbar=self.vscrollbar,
            hscrollbar=self.hscrollbar,
            width=self.width,
            height=self.height,
            **kwargs_text,
        )
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        event = Event(self, window, self.key, EventType.KeyRelease)
        self.widget.bind("<KeyRelease>", window._make_callback(event))

        self.value = self._value

        if self._command:
            self.events = True
            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.KeyRelease,
                    command=self._command,
                )
            )

        if self._disabled:
            self.widget["state"] = "disabled"

        return self.widget

    @property
    def value(self):
        return self.widget.get("1.0", tk.END).rstrip()

    @value.setter
    def value(self, text):
        self.widget.delete("1.0", tk.END)
        self.widget.insert("1.0", text)

    @property
    def text(self):
        """Return the Tk text widget"""
        return self.widget


# TODO: how to make Output read-only?
class Output(Text):
    """Text box that redirects stderr and/or stdout to the text box."""

    def __init__(
        self,
        text: str | None = None,
        key: Hashable | None = None,
        width: int = 40,
        height: int = 20,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = False,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        vscrollbar: bool = True,
        hscrollbar: bool = False,
        stdout: bool = True,
        stderr: bool = True,
        echo: bool = False,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        """
        Initialize an Output widget.

        Args:
            text (str | None, optional): Default text for the text box. Defaults to None.
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            width (int, optional): Width of the text box. Defaults to 40.
            height (int, optional): Height of the text box. Defaults to 20.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to False.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            command (CommandType | None, optional): Command callback. Defaults to None.
            vscrollbar (bool, optional): Show vertical scrollbar. Defaults to False.
            hscrollbar (bool, optional): Show horizontal scrollbar. Defaults to False.
            stdout (bool, optional): Redirect stdout to the text box. Defaults to True.
            stderr (bool, optional): Redirect stderr to the text box. Defaults to True.
            echo (bool, optional): Echo stdout and stderr to the console. Defaults to False.
            weightx (int | None, optional): Weight of the widget in the x direction. Defaults to None.
            weighty (int | None, optional): Weight of the widget in the y direction. Defaults to None.
            focus (bool, optional): If True, widget has focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments are passed to tk Text.
        """
        super().__init__(
            text=text,
            key=key,
            width=width,
            height=height,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            vscrollbar=vscrollbar,
            hscrollbar=hscrollbar,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
            **kwargs,
        )

        self.kwargs = kwargs
        self._echo = echo
        self._stdout = stdout
        self._stderr = stderr

        # stores state for stdout and stderr redirection
        self._redirect = []
        self._redirect_id = {}

    def _create_widget(self, parent, window: "Window", row, col):
        self.widget = super()._create_widget(parent, window, row, col)

        # Unbind <KeyRelease> since this isn't for user input
        self.widget.unbind("<KeyRelease>")

        if self.events:
            event = Event(self, window, self.key, EventType.OutputWrite)
            self.window.root.bind_all(
                EventType.OutputWrite.value, window._make_callback(event)
            )

        self._configure_redirect()
        self.enable_redirect()

        return self.widget

    def _configure_redirect(self):
        """Configure stdout and stderr redirection."""
        if self._stdout:
            self._redirect.append(StdOutRedirect())
        if self._stderr:
            self._redirect.append(StdErrRedirect())
        for r in self._redirect:
            r.echo = self._echo
            self._redirect_id[r] = r.register(self._write)

    def _write(self, line):
        with contextlib.suppress(tk.TclError):
            # ignore TclError if widget has been destroyed while trying to write
            self.text.insert(tk.END, line)
            self.text.yview(tk.END)
        self.window.root.event_generate(EventType.OutputWrite.value)

    @property
    def echo(self):
        """Return True if stdout and stderr are echoed to the console."""
        return self._echo

    @echo.setter
    def echo(self, echo):
        """Set whether stdout and stderr are echoed to the console."""
        self._echo = echo
        for r in self._redirect:
            r.echo = echo

    def disable_redirect(self):
        """Disable redirecting stdout and stderr to the text box."""
        for r in self._redirect:
            r.disable_redirect()

    def enable_redirect(self):
        """Enable redirecting stdout and stderr to the text box."""
        for r in self._redirect:
            r.enable_redirect()

    def __del__(self):
        for r, id_ in self._redirect_id.items():
            r.deregister(id_)



"""Widget wrapper class to use custom widgets with GUITk"""

from __future__ import annotations

import tkinter as tk
from typing import TYPE_CHECKING, Any, Hashable

from .basewidget import BaseWidget
from .events import Event, EventCommand, EventType
from .types import CommandType, PadType, TooltipType

if TYPE_CHECKING:
    from .window import Window


def widget_class_factory(
    widget_class: type[tk.BaseWidget],
    value_type: tk.Variable | None = None,
    value_name: str | None = None,
    event_type: Any = None,
) -> type[BaseWidget]:
    """Create a Widget class for a custom widget.

    Args:
        widget_class (type[tk.BaseWidget]): Tkinter widget class.
        value_type (tk.Variable | None): Type of value to get/set from widget. For example, tk.StringVar.
        value_name (str): Name of widget bound value variable, if any.
            For example, 'textvariable' for tk.Entry or 'variable' for tk.Checkbutton.
        event_type (Any): Event type to use for this widget if the widget should generate events.
            Defaults to None.

    Returns:
        Widget class (type[guitk.BaseWidget]) for the custom widget.

    Note:
        If you widget supports bound values, for example, the `textvariable` argument of tk.Entry,
        you should pass the `value_type` and `value_name` arguments to this constructor which will
        ensure that the Widget class's `value` property is bound to the widget's value. If one of
        these arguments is specified, both must be specified.
    """

    if value_type and not value_name or value_name and not value_type:
        raise ValueError(
            "value_type and value_name must both be specified or both be None"
        )

    # _Widget class for custom widgets
    # defined here as a closure so it has access to widget_class, value_type, and value_name

    class _Widget(BaseWidget):
        """Widget wrapper class to use custom widgets with GUITk"""

        def __init__(
            self,
            key: Hashable | None = None,
            disabled: bool = False,
            rowspan: int | None = None,
            columnspan: int | None = None,
            padx: PadType | None = None,
            pady: PadType | None = None,
            events: bool = True,
            sticky: str | None = None,
            tooltip: TooltipType = None,
            command: CommandType | None = None,
            weightx: int | None = None,
            weighty: int | None = None,
            focus: bool = False,
            **kwargs: Any,
        ):
            """Initialize a widget.

            Args:
                widget_class (type): Tkinter widget class.
                value_type: Type of value to get/set from widget. For example, tk.StringVar.
                value_name (str, optional): Name of widget bound value variable, if any.
                    For example, 'textvariable' for tk.Entry or 'variable' for tk.Checkbutton.
                key (Hashable, optional): Unique key for this widget. Defaults to None.
                default (str | None, optional): Default text for the entry box. Defaults to None.
                disabled (bool, optional): If True, widget is disabled. Defaults to False.
                columnspan (int | None, optional): Number of columns to span. Defaults to None.
                rowspan (int | None, optional): Number of rows to span. Defaults to None.
                padx (PadType | None, optional): X padding. Defaults to None.
                pady (PadType | None, optional): Y padding. Defaults to None.
                events (bool, optional): Enable events for this widget. Defaults to False.
                sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
                tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
                command (CommandType | None, optional): Command callback. Defaults to None.
                weightx (int | None, optional): Weight for horizontal resizing. Defaults to None.
                weighty (int | None, optional): Weight for vertical resizing. Defaults to None.
                focus (bool, optional): If True, widget has focus. Defaults to False.
                    Only one widget in a window can have focus.HLayout
                **kwargs: Additional keyword arguments are passed to widget_class.__init__().

            Note:
                If you widget supports bound values, for example, the `textvariable` argument of tk.Entry,
                you should pass the `value_type` and `value_name` arguments to this constructor which will
                ensure that the Widget class's `value` property is bound to the widget's value. If one of
                these arguments is specified, both must be specified.
            """

            super().__init__(
                key=key,
                disabled=disabled,
                rowspan=rowspan,
                columnspan=columnspan,
                padx=padx,
                pady=pady,
                events=events,
                sticky=sticky,
                tooltip=tooltip,
                command=command,
                weightx=weightx,
                weighty=weighty,
                focus=focus,
                value_type=value_type,
            )
            self.widget_class = widget_class
            self.kwargs = kwargs
            self.event_type = event_type
            self.value_name = value_name

        def _create_widget(self, parent: Any, window: Window, row: int, col: int):
            """Create the widget."""

            if self.event_type or self._command:
                self.event_type = self.event_type or f"{self.widget_class}"
                event = Event(self, window, self.key, self.event_type)
                command = window._make_callback(event)
                self.kwargs["command"] = command

            if self.value_name:
                self.kwargs[self.value_name] = self._value

            self.widget = self.widget_class(
                parent,
                **self.kwargs,
            )
            self._grid(
                row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
            )

            if self._command:
                self.events = True
                window._bind_command(
                    EventCommand(
                        widget=self,
                        key=self.key,
                        event_type=self.event_type,
                        command=self._command,
                    )
                )

            if self._disabled:
                self.widget.state(["disabled"])

            return self.widget

    return _Widget


class Widget(BaseWidget):
    """Widget wrapper class to use custom widgets with GUITk"""

    # use widget_class_Factory to create a Widget class for a custom widget in the __new__ method
    # so that the widget_class, value_type, and value_name arguments are available to the _Widget class
    # defined in the widget_class_Factory closure
    def __new__(cls, **kwargs: Any):
        """Create a Widget class for a custom widget.

        Args:
            widget_class (type[tk.BaseWidget]): Tkinter widget class.
            value_type (tk.Variable | None): Type of value to get/set from widget. For example, tk.StringVar.
            value_name (str): Name of widget bound value variable, if any.
                For example, 'textvariable' for tk.Entry or 'variable' for tk.Checkbutton.
            event_type (Any): Event type to use for this widget if the widget should generate events.
                Defaults to None.

        Returns:
            Widget class (type[guitk.BaseWidget]) for the custom widget.

        Note:
            If you widget supports bound values, for example, the `textvariable` argument of tk.Entry,
            you should pass the `value_type` and `value_name` arguments to this constructor which will
            ensure that the Widget class's `value` property is bound to the widget's value. If one of
            these arguments is specified, both must be specified.
        """
        widget_class = kwargs.pop("widget_class")
        value_type = kwargs.pop("value_type", None)
        value_name = kwargs.pop("value_name", None)
        event_type = kwargs.pop("event_type", None)

        _Widget = widget_class_factory(widget_class, value_type, value_name, event_type)

        return _Widget(**kwargs)

    def __init__(
        self,
        widget_class: type[tk.BaseWidget],
        value_type: tk.Variable | None = None,
        value_name: str | None = None,
        event_type: Any = None,
        key: Hashable | None = None,
        disabled: bool = False,
        rowspan: int | None = None,
        columnspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        command: CommandType | None = None,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs: Any,
    ):
        # This method is only needed to satisfy the type checker and IDE autocomplete
        # The actual Widget class is created in the __new__ method
        ...
